#!/usr/bin/python2


# Copyright (C) 2016 Sixten Bergman
# License WTFPL
#
# This program is free software. It comes without any warranty, to the extent
# permitted by applicable law. 
# You can redistribute it and/or modify it under the terms of the Do What The
# Fuck You Want To Public License, Version 2, as published by Sam Hocevar. See
# http://www.wtfpl.net/ for more details.
#
# note that the function peakdetect is derived from code which was released to
# public domain see: http://billauer.co.il/peakdet.html
#

from math import pi, log
import numpy as np
import pylab
from scipy import fft, ifft
from scipy.optimize import curve_fit
from scipy.signal import cspline1d_eval, cspline1d

__all__ = [
        "peakdetect",
        "peakdetect_fft",
        "peakdetect_parabola",
        "peakdetect_sine",
        "peakdetect_sine_locked",
        "peakdetect_spline",
        "peakdetect_zero_crossing",
        "zero_crossings",
        "zero_crossings_sine_fit"
        ]



def _datacheck_peakdetect(x_axis, y_axis):
    if x_axis is None:
        x_axis = range(len(y_axis))
    
    if len(y_axis) != len(x_axis):
        raise ValueError( 
                "Input vectors y_axis and x_axis must have same length")
    
    #needs to be a numpy array
    y_axis = np.array(y_axis)
    x_axis = np.array(x_axis)
    return x_axis, y_axis
    

def _pad(fft_data, pad_len):
    """
    Pads fft data to interpolate in time domain
    
    keyword arguments:
    fft_data -- the fft
    pad_len --  By how many times the time resolution should be increased by
    
    return: padded list
    """
    l = len(fft_data)
    n = _n(l * pad_len)
    fft_data = list(fft_data)
    
    return fft_data[:l // 2] + [0] * (2**n-l) + fft_data[l // 2:]
    
def _n(x):
    """
    Find the smallest value for n, which fulfils 2**n >= x
    
    keyword arguments:
    x -- the value, which 2**n must surpass
    
    return: the integer n
    """
    return int(log(x)/log(2)) + 1
    
    
def _peakdetect_parabola_fitter(raw_peaks, x_axis, y_axis, points):
    """
    Performs the actual parabola fitting for the peakdetect_parabola function.
        
    keyword arguments:
    raw_peaks -- A list of either the maxima or the minima peaks, as given
        by the peakdetect functions, with index used as x-axis
    
    x_axis -- A numpy array of all the x values
    
    y_axis -- A numpy array of all the y values
    
    points -- How many points around the peak should be used during curve
        fitting, must be odd.
    
    
    return: A list giving all the peaks and the fitted waveform, format:
        [[x, y, [fitted_x, fitted_y]]]
        
    """
    func = lambda x, a, tau, c: a * ((x - tau) ** 2) + c
    fitted_peaks = []
    distance = abs(x_axis[raw_peaks[1][0]] - x_axis[raw_peaks[0][0]]) / 4
    for peak in raw_peaks:
        index = peak[0]
        x_data = x_axis[index - points // 2: index + points // 2 + 1]
        y_data = y_axis[index - points // 2: index + points // 2 + 1]
        # get a first approximation of tau (peak position in time)
        tau = x_axis[index]
        # get a first approximation of peak amplitude
        c = peak[1]
        a = np.sign(c) * (-1) * (np.sqrt(abs(c))/distance)**2
        """Derived from ABC formula to result in a solution where A=(rot(c)/t)**2"""
        
        # build list of approximations
        
        p0 = (a, tau, c)
        popt, pcov = curve_fit(func, x_data, y_data, p0)
        # retrieve tau and c i.e x and y value of peak
        x, y = popt[1:3]
        
        # create a high resolution data set for the fitted waveform
        x2 = np.linspace(x_data[0], x_data[-1], points * 10)
        y2 = func(x2, *popt)
        
        fitted_peaks.append([x, y, [x2, y2]])
        
    return fitted_peaks
    
    
    
def peakdetect(y_axis, x_axis = None, lookahead = 200, delta=0):
    """
    Converted from/based on a MATLAB script at: 
    http://billauer.co.il/peakdet.html
    
    function for detecting local maxima and minima in a signal.
    Discovers peaks by searching for values which are surrounded by lower
    or larger values for maxima and minima respectively
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks. If omitted an
        index of the y_axis is used.
        (default: None)
    
    lookahead -- distance to look ahead from a peak candidate to determine if
        it is the actual peak
        (default: 200) 
        '(samples / period) / f' where '4 >= f >= 1.25' might be a good value
    
    delta -- this specifies a minimum difference between a peak and
        the following points, before a peak may be considered a peak. Useful
        to hinder the function from picking up false peaks towards to end of
        the signal. To work well delta should be set to delta >= RMSnoise * 5.
        (default: 0)
            When omitted delta function causes a 20% decrease in speed.
            When used Correctly it can double the speed of the function
    
    
    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*max_peaks)
    """
    max_peaks = []
    min_peaks = []
    dump = []   #Used to pop the first hit which almost always is false
       
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # store data length for later use
    length = len(y_axis)
    
    
    #perform some checks
    if lookahead < 1:
        raise ValueError("Lookahead must be '1' or above in value")
    if not (np.isscalar(delta) and delta >= 0):
        raise ValueError("delta must be a positive number")
    
    #maxima and minima candidates are temporarily stored in
    #mx and mn respectively
    mn, mx = np.Inf, -np.Inf
    
    #Only detect peak if there is 'lookahead' amount of points after it
    for index, (x, y) in enumerate(zip(x_axis[:-lookahead], 
                                        y_axis[:-lookahead])):
        if y > mx:
            mx = y
            mxpos = x
        if y < mn:
            mn = y
            mnpos = x
        
        ####look for max####
        if y < mx-delta and mx != np.Inf:
            #Maxima peak candidate found
            #look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index+lookahead].max() < mx:
                max_peaks.append([mxpos, mx])
                dump.append(True)
                #set algorithm to only find minima now
                mx = np.Inf
                mn = np.Inf
                if index+lookahead >= length:
                    #end is within lookahead no more peaks can be found
                    break
                continue
            #else:  #slows shit down this does
            #    mx = ahead
            #    mxpos = x_axis[np.where(y_axis[index:index+lookahead]==mx)]
        
        ####look for min####
        if y > mn+delta and mn != -np.Inf:
            #Minima peak candidate found 
            #look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index+lookahead].min() > mn:
                min_peaks.append([mnpos, mn])
                dump.append(False)
                #set algorithm to only find maxima now
                mn = -np.Inf
                mx = -np.Inf
                if index+lookahead >= length:
                    #end is within lookahead no more peaks can be found
                    break
            #else:  #slows shit down this does
            #    mn = ahead
            #    mnpos = x_axis[np.where(y_axis[index:index+lookahead]==mn)]
    
    
    #Remove the false hit on the first value of the y_axis
    try:
        if dump[0]:
            max_peaks.pop(0)
        else:
            min_peaks.pop(0)
        del dump
    except IndexError:
        #no peaks were found, should the function return empty lists?
        pass
        
    return [max_peaks, min_peaks]
    
    
def peakdetect_fft(y_axis, x_axis, pad_len = 20):
    """
    Performs a FFT calculation on the data and zero-pads the results to
    increase the time domain resolution after performing the inverse fft and
    send the data to the 'peakdetect' function for peak 
    detection.
    
    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as the index 50.234 or similar.
    
    Will find at least 1 less peak then the 'peakdetect_zero_crossing'
    function, but should result in a more precise value of the peak as
    resolution has been increased. Some peaks are lost in an attempt to
    minimize spectral leakage by calculating the fft between two zero
    crossings for n amount of signal periods.
    
    The biggest time eater in this function is the ifft and thereafter it's
    the 'peakdetect' function which takes only half the time of the ifft.
    Speed improvements could include to check if 2**n points could be used for
    fft and ifft or change the 'peakdetect' to the 'peakdetect_zero_crossing',
    which is maybe 10 times faster than 'peakdetct'. The pro of 'peakdetect'
    is that it results in one less lost peak. It should also be noted that the
    time used by the ifft function can change greatly depending on the input.
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    
    pad_len -- By how many times the time resolution should be
        increased by, e.g. 1 doubles the resolution. The amount is rounded up
        to the nearest 2**n amount
        (default: 20)
    
    
    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    zero_indices = zero_crossings(y_axis, window_len = 11)
    #select a n amount of periods
    last_indice = - 1 - (1 - len(zero_indices) & 1)
    ###
    # Calculate the fft between the first and last zero crossing
    # this method could be ignored if the beginning and the end of the signal
    # are unnecessary as any errors induced from not using whole periods
    # should mainly manifest in the beginning and the end of the signal, but
    # not in the rest of the signal
    # this is also unnecessary if the given data is an amount of whole periods
    ###
    fft_data = fft(y_axis[zero_indices[0]:zero_indices[last_indice]])
    padd = lambda x, c: x[:len(x) // 2] + [0] * c + x[len(x) // 2:]
    n = lambda x: int(log(x)/log(2)) + 1
    # pads to 2**n amount of samples
    fft_padded = padd(list(fft_data), 2 ** 
                n(len(fft_data) * pad_len) - len(fft_data))
    
    # There is amplitude decrease directly proportional to the sample increase
    sf = len(fft_padded) / float(len(fft_data))
    # There might be a leakage giving the result an imaginary component
    # Return only the real component
    y_axis_ifft = ifft(fft_padded).real * sf #(pad_len + 1)
    x_axis_ifft = np.linspace(
                x_axis[zero_indices[0]], x_axis[zero_indices[last_indice]],
                len(y_axis_ifft))
    # get the peaks to the interpolated waveform
    max_peaks, min_peaks = peakdetect(y_axis_ifft, x_axis_ifft, 500,
                                    delta = abs(np.diff(y_axis).max() * 2))
    #max_peaks, min_peaks = peakdetect_zero_crossing(y_axis_ifft, x_axis_ifft)
    
    # store one 20th of a period as waveform data
    data_len = int(np.diff(zero_indices).mean()) / 10
    data_len += 1 - data_len & 1
    
    
    return [max_peaks, min_peaks]
    
    
def peakdetect_parabola(y_axis, x_axis, points = 31):
    """
    Function for detecting local maxima and minima in a signal.
    Discovers peaks by fitting the model function: y = k (x - tau) ** 2 + m
    to the peaks. The amount of points used in the fitting is set by the
    points argument.
    
    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly, if it was returned as index 50.234 or similar.
    
    will find the same amount of peaks as the 'peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    
    points -- How many points around the peak should be used during curve
        fitting (default: 31)
    
    
    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # make the points argument odd
    points += 1 - points % 2
    #points += 1 - int(points) & 1 slower when int conversion needed
    
    # get raw peaks
    max_raw, min_raw = peakdetect_zero_crossing(y_axis)
    
    # define output variable
    max_peaks = []
    min_peaks = []
    
    max_ = _peakdetect_parabola_fitter(max_raw, x_axis, y_axis, points)
    min_ = _peakdetect_parabola_fitter(min_raw, x_axis, y_axis, points)
    
    max_peaks = map(lambda x: [x[0], x[1]], max_)
    min_peaks = map(lambda x: [x[0], x[1]], min_)
    
    return [max_peaks, min_peaks]
    

def peakdetect_sine(y_axis, x_axis, points = 31, lock_frequency = False):
    """
    Function for detecting local maxima and minima in a signal.
    Discovers peaks by fitting the model function:
    y = A * sin(2 * pi * f * (x - tau)) to the peaks. The amount of points used
    in the fitting is set by the points argument.
    
    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as index 50.234 or similar.
    
    will find the same amount of peaks as the 'peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.
    
    The function might have some problems if the sine wave has a
    non-negligible total angle i.e. a k*x component, as this messes with the
    internal offset calculation of the peaks, might be fixed by fitting a 
    y = k * x + m function to the peaks for offset calculation.
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    
    points -- How many points around the peak should be used during curve
        fitting (default: 31)
    
    lock_frequency -- Specifies if the frequency argument of the model
        function should be locked to the value calculated from the raw peaks
        or if optimization process may tinker with it.
        (default: False)
    
    
    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # make the points argument odd
    points += 1 - points % 2
    #points += 1 - int(points) & 1 slower when int conversion needed
    
    # get raw peaks
    max_raw, min_raw = peakdetect_zero_crossing(y_axis)
    
    # define output variable
    max_peaks = []
    min_peaks = []
    
    # get global offset
    offset = np.mean([np.mean(max_raw, 0)[1], np.mean(min_raw, 0)[1]])
    # fitting a k * x + m function to the peaks might be better
    #offset_func = lambda x, k, m: k * x + m
    
    # calculate an approximate frequency of the signal
    Hz_h_peak = np.diff(zip(*max_raw)[0]).mean()
    Hz_l_peak = np.diff(zip(*min_raw)[0]).mean()
    Hz = 1 / np.mean([Hz_h_peak, Hz_l_peak])
    
    
    
    # model function
    # if cosine is used then tau could equal the x position of the peak
    # if sine were to be used then tau would be the first zero crossing
    if lock_frequency:
        func = lambda x_ax, A, tau: A * np.sin(
            2 * pi * Hz * (x_ax - tau) + pi / 2)
    else:
        func = lambda x_ax, A, Hz, tau: A * np.sin(
            2 * pi * Hz * (x_ax - tau) + pi / 2)
    #func = lambda x_ax, A, Hz, tau: A * np.cos(2 * pi * Hz * (x_ax - tau))
    
    
    #get peaks
    fitted_peaks = []
    for raw_peaks in [max_raw, min_raw]:
        peak_data = []
        for peak in raw_peaks:
            index = peak[0]
            x_data = x_axis[index - points // 2: index + points // 2 + 1]
            y_data = y_axis[index - points // 2: index + points // 2 + 1]
            # get a first approximation of tau (peak position in time)
            tau = x_axis[index]
            # get a first approximation of peak amplitude
            A = peak[1]
            
            # build list of approximations
            if lock_frequency:
                p0 = (A, tau)
            else:
                p0 = (A, Hz, tau)
            
            # subtract offset from wave-shape
            y_data -= offset
            popt, pcov = curve_fit(func, x_data, y_data, p0)
            # retrieve tau and A i.e x and y value of peak
            x = popt[-1]
            y = popt[0]
            
            # create a high resolution data set for the fitted waveform
            x2 = np.linspace(x_data[0], x_data[-1], points * 10)
            y2 = func(x2, *popt)
            
            # add the offset to the results
            y += offset
            y2 += offset
            y_data += offset
            
            peak_data.append([x, y, [x2, y2]])
       
        fitted_peaks.append(peak_data)
    
    # structure date for output
    max_peaks = map(lambda x: [x[0], x[1]], fitted_peaks[0])
    min_peaks = map(lambda x: [x[0], x[1]], fitted_peaks[1])
    
    
    return [max_peaks, min_peaks]

    
def peakdetect_sine_locked(y_axis, x_axis, points = 31):
    """
    Convenience function for calling the 'peakdetect_sine' function with
    the lock_frequency argument as True.
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    points -- How many points around the peak should be used during curve
        fitting (default: 31)
    
    return: see the function 'peakdetect_sine'
    """
    return peakdetect_sine(y_axis, x_axis, points, True)
    
    
def peakdetect_spline(y_axis, x_axis, pad_len=20):
    """
    Performs a b-spline interpolation on the data to increase resolution and
    send the data to the 'peakdetect_zero_crossing' function for peak 
    detection.
    
    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as the index 50.234 or similar.
    
    will find the same amount of peaks as the 'peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks. 
        x-axis must be equally spaced.
    
    pad_len -- By how many times the time resolution should be increased by,
        e.g. 1 doubles the resolution.
        (default: 20)
    
    
    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # could perform a check if x_axis is equally spaced
    #if np.std(np.diff(x_axis)) > 1e-15: raise ValueError
    # perform spline interpolations
    dx = x_axis[1] - x_axis[0]
    x_interpolated = np.linspace(x_axis.min(), x_axis.max(), len(x_axis) * (pad_len + 1))
    cj = cspline1d(y_axis)
    y_interpolated = cspline1d_eval(cj, x_interpolated, dx=dx,x0=x_axis[0])
    # get peaks
    max_peaks, min_peaks = peakdetect_zero_crossing(y_interpolated, x_interpolated)
    
    return [max_peaks, min_peaks]
    
def peakdetect_zero_crossing(y_axis, x_axis = None, window = 11):
    """
    Function for detecting local maxima and minima in a signal.
    Discovers peaks by dividing the signal into bins and retrieving the
    maximum and minimum value of each the even and odd bins respectively.
    Division into bins is performed by smoothing the curve and finding the
    zero crossings.
    
    Suitable for repeatable signals, where some noise is tolerated. Executes
    faster than 'peakdetect', although this function will break if the offset
    of the signal is too large. It should also be noted that the first and
    last peak will probably not be found, as this function only can find peaks
    between the first and last zero crossing.
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    
    x_axis -- A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the position of the peaks. If
        omitted an index of the y_axis is used.
        (default: None)
    
    window -- the dimension of the smoothing window; should be an odd integer
        (default: 11)
    
    
    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value) 
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do: 
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    
    zero_indices = zero_crossings(y_axis, window_len = window)
    period_lengths = np.diff(zero_indices)
            
    bins_y = [y_axis[index:index + diff] for index, diff in 
        zip(zero_indices, period_lengths)]
    bins_x = [x_axis[index:index + diff] for index, diff in 
        zip(zero_indices, period_lengths)]
        
    even_bins_y = bins_y[::2]
    odd_bins_y = bins_y[1::2]
    even_bins_x = bins_x[::2]
    odd_bins_x = bins_x[1::2]
    hi_peaks_x = []
    lo_peaks_x = []
    
    #check if even bin contains maxima
    if abs(even_bins_y[0].max()) > abs(even_bins_y[0].min()):
        hi_peaks = [bin.max() for bin in even_bins_y]
        lo_peaks = [bin.min() for bin in odd_bins_y]
        # get x values for peak
        for bin_x, bin_y, peak in zip(even_bins_x, even_bins_y, hi_peaks):
            hi_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
        for bin_x, bin_y, peak in zip(odd_bins_x, odd_bins_y, lo_peaks):
            lo_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
    else:
        hi_peaks = [bin.max() for bin in odd_bins_y]
        lo_peaks = [bin.min() for bin in even_bins_y]
        # get x values for peak
        for bin_x, bin_y, peak in zip(odd_bins_x, odd_bins_y, hi_peaks):
            hi_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
        for bin_x, bin_y, peak in zip(even_bins_x, even_bins_y, lo_peaks):
            lo_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
    
    max_peaks = [[x, y] for x,y in zip(hi_peaks_x, hi_peaks)]
    min_peaks = [[x, y] for x,y in zip(lo_peaks_x, lo_peaks)]
    
    return [max_peaks, min_peaks]
        
    
def _smooth(x, window_len=11, window="hanning"):
    """
    smooth the data using a window of the requested size.
    
    This method is based on the convolution of a scaled window on the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the beginning and end part of the output signal.
    
    keyword arguments:
    x -- the input signal 
    
    window_len -- the dimension of the smoothing window; should be an odd
        integer (default: 11)
    
    window -- the type of window from 'flat', 'hanning', 'hamming', 
        'bartlett', 'blackman', where flat is a moving average
        (default: 'hanning')

    
    return: the smoothed signal
        
    example:

    t = linspace(-2,2,0.1)
    x = sin(t)+randn(len(t))*0.1
    y = _smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, 
    numpy.convolve, scipy.signal.lfilter 
    """
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")
    
    if window_len<3:
        return x
    #declare valid windows in a dictionary
    window_funcs = {
        "flat": lambda _len: np.ones(_len, "d"),
        "hanning": np.hanning,
        "hamming": np.hamming,
        "bartlett": np.bartlett,
        "blackman": np.blackman
        }
    
    s = np.r_[x[window_len-1:0:-1], x, x[-1:-window_len:-1]]
    try:
        w = window_funcs[window](window_len)
    except KeyError:
        raise ValueError(
            "Window is not one of '{0}', '{1}', '{2}', '{3}', '{4}'".format(
            *window_funcs.keys()))
    
    y = np.convolve(w / w.sum(), s, mode = "valid")
    
    return y
    
    
def zero_crossings(y_axis, window_len = 11, 
    window_f="hanning", offset_corrected=False):
    """
    Algorithm to find zero crossings. Smooths the curve and finds the
    zero-crossings by looking for a sign change.
    
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find zero-crossings
    
    window_len -- the dimension of the smoothing window; should be an odd
        integer (default: 11)
    
    window_f -- the type of window from 'flat', 'hanning', 'hamming', 
        'bartlett', 'blackman' (default: 'hanning')
    
    offset_corrected -- Used for recursive calling to remove offset when needed
    
    
    return: the index for each zero-crossing
    """
    # smooth the curve
    length = len(y_axis)
    
    # discard tail of smoothed signal
    y_axis = _smooth(y_axis, window_len, window_f)[:length]
    indices = np.where(np.diff(np.sign(y_axis)))[0]
    
    # check if zero-crossings are valid
    diff = np.diff(indices)
    if diff.std() / diff.mean() > 0.1:
        #Possibly bad zero crossing, see if it's offsets
        if ((diff[::2].std() / diff[::2].mean()) < 0.1 and 
        (diff[1::2].std() / diff[1::2].mean()) < 0.1 and
        not offset_corrected):
            #offset present attempt to correct by subtracting the average
            offset = np.mean([y_axis.max(), y_axis.min()])
            return zero_crossings(y_axis-offset, window_len, window_f, True)
        #Invalid zero crossings and the offset has been removed
        print(diff.std() / diff.mean())
        print(np.diff(indices))
        raise ValueError(
            "False zero-crossings found, indicates problem {0!s} or {1!s}".format(
            "with smoothing window", "unhandled problem with offset"))
    # check if any zero crossings were found
    if len(indices) < 1:
        raise ValueError("No zero crossings found")
    #remove offset from indices due to filter function when returning
    return indices - (window_len // 2 - 1)
    # used this to test the fft function's sensitivity to spectral leakage
    #return indices + np.asarray(30 * np.random.randn(len(indices)), int)
    
############################Frequency calculation#############################
#    diff = np.diff(indices)
#    time_p_period = diff.mean()
#    
#    if diff.std() / time_p_period > 0.1:
#        raise ValueError(
#            "smoothing window too small, false zero-crossing found")
#    
#    #return frequency
#    return 1.0 / time_p_period
##############################################################################


def zero_crossings_sine_fit(y_axis, x_axis, fit_window = None, smooth_window = 11):
    """
    Detects the zero crossings of a signal by fitting a sine model function
    around the zero crossings:
    y = A * sin(2 * pi * Hz * (x - tau)) + k * x + m
    Only tau (the zero crossing) is varied during fitting.
    
    Offset and a linear drift of offset is accounted for by fitting a linear
    function the negative respective positive raw peaks of the wave-shape and
    the amplitude is calculated using data from the offset calculation i.e.
    the 'm' constant from the negative peaks is subtracted from the positive
    one to obtain amplitude.
    
    Frequency is calculated using the mean time between raw peaks.
    
    Algorithm seems to be sensitive to first guess e.g. a large smooth_window
    will give an error in the results.
    
    
    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    
    x_axis -- A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the position of the peaks. If
        omitted an index of the y_axis is used. (default: None)
    
    fit_window -- Number of points around the approximate zero crossing that
        should be used when fitting the sine wave. Must be small enough that
        no other zero crossing will be seen. If set to none then the mean
        distance between zero crossings will be used (default: None)
    
    smooth_window -- the dimension of the smoothing window; should be an odd
        integer (default: 11)
    
    
    return: A list containing the positions of all the zero crossings.
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    #get first guess
    zero_indices = zero_crossings(y_axis, window_len = smooth_window)
    #modify fit_window to show distance per direction
    if fit_window is None:
        fit_window = np.diff(zero_indices).mean() // 3
    else:
        fit_window = fit_window // 2
    
    #x_axis is a np array, use the indices to get a subset with zero crossings
    approx_crossings = x_axis[zero_indices]
    
    
    
    #get raw peaks for calculation of offsets and frequency
    raw_peaks = peakdetect_zero_crossing(y_axis, x_axis)
    #Use mean time between peaks for frequency
    ext = lambda x: list(zip(*x)[0])
    _diff = map(np.diff, map(ext, raw_peaks))
    
    
    Hz = 1 / np.mean(map(np.mean, _diff))
    #Hz = 1 / np.diff(approx_crossings).mean() #probably bad precision
    
    
    #offset model function
    offset_func = lambda x, k, m: k * x + m
    k = []
    m = []
    amplitude = []
    
    for peaks in raw_peaks:
        #get peak data as nparray
        x_data, y_data = map(np.asarray, zip(*peaks))
        #x_data = np.asarray(x_data)
        #y_data = np.asarray(y_data)
        #calc first guess
        A = np.mean(y_data)
        p0 = (0, A)
        popt, pcov = curve_fit(offset_func, x_data, y_data, p0)
        #append results
        k.append(popt[0])
        m.append(popt[1])
        amplitude.append(abs(A))
    
    #store offset constants
    p_offset = (np.mean(k), np.mean(m))
    A = m[0] - m[1]
    #define model function to fit to zero crossing
    #y = A * sin(2*pi * Hz * (x - tau)) + k * x + m
    func = lambda x, tau: A * np.sin(2 * pi * Hz * (x - tau)) + offset_func(x, *p_offset)
    
    
    #get true crossings
    true_crossings = []
    for indice, crossing in zip(zero_indices, approx_crossings):
        p0 = (crossing, )
        subset_start = max(indice - fit_window, 0.0)
        subset_end = min(indice + fit_window + 1, len(x_axis) - 1.0)
        x_subset = np.asarray(x_axis[subset_start:subset_end])
        y_subset = np.asarray(y_axis[subset_start:subset_end])
        #fit
        popt, pcov = curve_fit(func, x_subset, y_subset, p0)
        
        true_crossings.append(popt[0])
    
    
    return true_crossings
        
        
    
    
def _test_zero():
    _max, _min = peakdetect_zero_crossing(y,x)
def _test():
    _max, _min = peakdetect(y,x, delta=0.30)
    
    
def _test_graph():
    i = 10000
    x = np.linspace(0,3.7*pi,i)
    y = (0.3*np.sin(x) + np.sin(1.3 * x) + 0.9 * np.sin(4.2 * x) + 0.06 *
    np.random.randn(i))
    y *= -1
    x = range(i)
    
    _max, _min = peakdetect(y,x,750, 0.30)
    xm = [p[0] for p in _max]
    ym = [p[1] for p in _max]
    xn = [p[0] for p in _min]
    yn = [p[1] for p in _min]
    
    pylab.plot(x,y)
    pylab.hold(True)
    pylab.plot(xm, ym, "r+")
    pylab.plot(xn, yn, "g+")
    
    pylab.show()
    
def _test_graph_cross(window = 11):
    i = 10000
    x = np.linspace(0,8.7*pi,i)
    y = (2*np.sin(x) + 0.006 *
    np.random.randn(i))
    y *= -1
    pylab.plot(x,y)
    #pylab.show()
    
    
    crossings = zero_crossings_sine_fit(y,x, smooth_window = window)
    y_cross = [0] * len(crossings)
    
    
    pylab.plot(x,y)
    pylab.hold(True)
    pylab.plot(crossings, y_cross, "b+")
    pylab.show()
    
    
    
if __name__ == "__main__":
    
    it = 10000
    xl = np.linspace(0,3.7*pi,it)
    yl = (0.3*np.sin(xl) + np.sin(1.3 * xl) + 0.9 * np.sin(4.2 * xl) + 0.06 * 
    np.random.randn(it))
    yl *= -1
    
    _max, _min = peakdetect(yl, xl, 750, 0.30)
    xm1 = [p[0] for p in _max]
    ym1 = [p[1] for p in _max]
    xn1 = [p[0] for p in _min]
    yn1 = [p[1] for p in _min]
    
    pylab.plot(xl, yl)
    pylab.hold(True)
    pylab.plot(xm1, ym1, "r+")
    pylab.plot(xn1, yn1, "g+")
    
    
    pylab.show()


import os
import sys

import numpy as np
import pandas as pd
import scipy.stats as st
import matplotlib.pyplot as plt
plt.rcParams["font.family"] = "Times New Roman"

PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(PATH, '..'))
sys.path.append(PATH)
from utils import save_dict_kaldimat, get_dict
from AEspeech import AEspeech
from script_mananger import script_manager
import torch
from tqdm import tqdm

class RepLearning:
    """
    Feature extraction from speech signals based on representation learning strategies using convolutional and recurrent autoencoders

    Two types of features are computed

    1. 256 features extracted from the bottleneck layer of the autoencoders
    2. 128 features based on the MSE between the decoded and input spectrograms of the autoencoder in different frequency regions


    Additionally, static (for all utterance) or dynamic (for each 500 ms speech segments) features can be computed:
    - The static feature vector is formed with 1024 features and contains (384 descriptors) x (4 functionals: mean, std, skewness, kurtosis)
    - The dynamic feature matrix is formed with the 384 descriptors computed for speech segments with 500ms length and 250ms time-shift
    - You can choose between features computed from a convolutional or recurrent autoencoder

    Script is called as follows

    >>> python replearning.py <file_or_folder_audio> <file_features> <static (true or false)> <plots (true or false)> <format (csv, txt, npy, kaldi, torch)> <model (CAE, RAE)>

    Examples command line:

    >>> python replearning.py "../audios/001_ddk1_PCGITA.wav" "replearningfeaturesDDKst.txt" "true" "true" "txt" "CAE"
    >>> python replearning.py "../audios/001_ddk1_PCGITA.wav" "replearningfeaturesDDKdyn.pt" "false" "true" "torch" "CAE"

    >>> python replearning.py "../audios/" "replearningfeaturesst.txt" "true" "false" "txt" "CAE"
    >>> python replearning.py "../audios/" "replearningfeaturesst.csv" "true" "false" "csv" "CAE"
    >>> python replearning.py "../audios/" "replearningfeaturesdyn.pt" "false" "false" "torch" "CAE"

    Examples directly in Python

    >>> from replearning import RepLearning
    >>> replearning=RepLearning('CAE')
    >>> file_audio="../audios/001_a1_PCGITA.wav"
    >>> features1=replearning.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
    >>> features2=replearning.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
    >>> features3=replearning.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
    >>> replearning.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")
    """

    def __init__(self, model):
        self.size_bottleneck=256
        self.AEspeech=AEspeech(model, self.size_bottleneck)
        self.statistics=["mean", "std", "skewness", "kurtosis"]

        feat_names_bottle=["bottleneck_"+str(k) for k in range(self.size_bottleneck)]
        feat_names_error=["error_"+str(k) for k in range(128)]
        feat_names_bottle_all=[]
        feat_names_error_all=[]

        for k in self.statistics:
            for j in feat_names_bottle:
                feat_names_bottle_all.append(k+"_"+j)
            for j in feat_names_error:
                feat_names_error_all.append(k+"_"+j)

        self.head_st=feat_names_bottle_all+feat_names_error_all
        self.head_dyn=np.hstack(feat_names_bottle+feat_names_error)


    def extract_features_file(self, audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """
        Extract the representation learning features from an audio file

        :param audio: .wav audio file.
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldi features, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> replearning=RepLearning('CAE')
        >>> file_audio="../audios/001_ddk1_PCGITA.wav"
        >>> features1=replearning.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
        >>> features2=replearning.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
        >>> features3=replearning.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
        >>> replearning.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")

        >>> replearning=RepLearning('CAE')
        >>> path_audio="../audios/"
        >>> features1=replearning.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=replearning.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=replearning.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
        >>> replearning.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")

        """
        if static and fmt=="kaldi":
            raise ValueError("Kaldi is only supported for dynamic features")
        hb=self.AEspeech.compute_bottleneck_features(audio)
        err=self.AEspeech.compute_rec_error_features(audio)
        if plots:
            self.AEspeech.plot_spectrograms(audio)
        if static:
            bottle_feat=np.hstack((np.mean(hb, 0), np.std(hb, 0), st.skew(hb, 0), st.kurtosis(hb, 0)))
            error_feat=np.hstack((np.mean(err, 0), np.std(err, 0), st.skew(err, 0), st.kurtosis(err, 0)))
            feat=np.hstack((bottle_feat, error_feat))
            feat=np.expand_dims(feat, axis=0)
            head=self.head_st
        else:
            feat=np.concatenate((hb, err), axis=1)
            head=self.head_dyn
        if fmt in("npy","txt"):
            return feat
        elif fmt in("dataframe","csv"):
            dff={}
            for e, key in enumerate(head):
                dff[key]=feat[:,e]
            dff=pd.DataFrame(dff)
            return dff
        elif fmt=="torch":
            return torch.from_numpy(feat)
        elif fmt=="kaldi":
            name_all=audio.split('/')
            dictX={name_all[-1]:feat}
            save_dict_kaldimat(dictX, kaldi_file)
        else:
            raise ValueError(fmt+" is not supported")

    def extract_features_path(self, path_audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """
        Extract the representation learning features for audios inside a path
        
        :param path_audio: directory with (.wav) audio files inside, sampled at 16 kHz
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldifeatures, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> replearning=RepLearning('CAE')
        >>> path_audio="../audios/"
        >>> features1=phonological.replearning(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=phonological.replearning(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=phonological.replearning(path_audio, static=False, plots=True, fmt="torch")
        >>> replearning.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")
        """
        if static and fmt=="kaldi":
            raise ValueError("Kaldi is only supported for dynamic features")
        hf=os.listdir(path_audio)
        hf.sort()

        pbar=tqdm(range(len(hf)))
        ids=[]

        Features=[]
        for j in pbar:
            pbar.set_description("Processing %s" % hf[j])
            audio_file=path_audio+hf[j]
            feat=self.extract_features_file(audio_file, static=static, plots=plots, fmt="npy")
            Features.append(feat)
            if static:
                ids.append(hf[j])
            else:
                ids.append(np.repeat(hf[j], feat.shape[0]))
        
        Features=np.vstack(Features)
        ids=np.hstack(ids)

        if static:
            head=self.head_st
        else:
            head=self.head_dyn

        if fmt in("npy","txt"):
            return Features
        if fmt in("dataframe","csv"):
            df={}
            for e, k in enumerate(head):
                df[k]=Features[:,e]
            df["id"]=ids
            return pd.DataFrame(df)
        if fmt=="torch":
            return torch.from_numpy(Features)
        if fmt=="kaldi":
            dictX=get_dict(Features, ids)
            save_dict_kaldimat(dictX, kaldi_file)
        else:
            raise ValueError(fmt+" is not supported")


if __name__=="__main__":

    if len(sys.argv)!=7:
        print("python replearning.py <file_or_folder_audio> <file_features> <static (true, false)> <plots (true,  false)> <format (csv, txt, npy, kaldi, torch)> <model (CAE,RAE)>")
        sys.exit()

    replearning=RepLearning(sys.argv[-1])
    script_manager(sys.argv, replearning)




# -*- coding: utf-8 -*-
"""
Created on Jul 21 2017

@author: J. C. Vasquez-Correa
"""


import numpy as np
import sys

def jitter_env(vPPS, iNumPuntos):

    iLongSec=len(vPPS)

    if (iLongSec < 2):
        sys.warn('Pitch sequence is too short' )
        vJitta=np.zeros(iNumPuntos)
        return vJitta

    vJitta=np.zeros(iNumPuntos)
    iIndiceIni=0
    iDesplazamiento=iLongSec/iNumPuntos
    rFoMed=np.max(vPPS)
    for n in range(iNumPuntos-1):
        indice=int( iIndiceIni+n*iDesplazamiento )
        if ( n>0 ) and (indice==int( iIndiceIni+(n-1)*iDesplazamiento )):
            vJitta[ n ]=vJitta[ n-1 ]
        else:
            if (indice+1 < iLongSec):
                    vJitta[n]=np.abs(vPPS[indice+1]-vPPS[indice])
            else:
                vJitta[n]=0
        vJitta[n]=100*vJitta[n]/rFoMed

    return vJitta

def get_log_energy(sig):
    sig2=np.asarray(sig)**2
    sumsig2=np.sum(np.absolute(sig2))/len(sig2)
    logE=np.log10(sumsig2)
    return logE


def get_power(sig):
    sig2=np.square(sig)
    sumsig2=np.sum(sig2)/len(sig)
    return sumsig2


def shimmer_env(vPPS, iNumPuntos):

    iLongSec=len(vPPS)

    if (iLongSec < 2):
        sys.warn( 'Sequence is too short' )
        vShimm=np.zeros(iNumPuntos)
        return vShimm

    vShimm=np.zeros(iNumPuntos)
    iIndiceIni=0

    iDesplazamiento=iLongSec/iNumPuntos
    rFoMed=np.max(vPPS)

    for n in range(iNumPuntos-1):
        indice=int( iIndiceIni+n*iDesplazamiento )
        if ( n>0 ) and (indice==int( iIndiceIni+(n-1)*iDesplazamiento )):
            vShimm[ n ]=vShimm[ n-1 ]
        else:
            if (indice+1 < iLongSec):
                    vShimm[n]=np.abs(vPPS[indice+1]-vPPS[indice])
            else:
                vShimm[n]=0
        vShimm[n]=100*vShimm[n]/rFoMed

    return vShimm



def PQ(x,k):
    """
    Perturbation Quotient in percentage of the signal x
    :params x: input sequence: F0 values or Amplitude values per window
    :params k: average factor (must be an odd number)
    :returns PQ
    """
    N=len(x)
    if N<k or k%2==0:
        return 0
    m=int(0.5*(k-1))
    summ=0
    for n in range(N-k):
        dif=0
        for r in range(k):
            dif=dif+x[n+r]-x[n+m]
        dif=np.abs(dif/float(k))
        summ=summ+dif
    num=summ/(N-k)
    den=np.mean(np.abs(x))
    c=100*num/den
    if np.sum(np.isnan(c))>0:
        print(x)
    return c

def APQ(PAS):
    """
    Amplitude perturbation quotient (APQ)
    :params PAS: sequence of amplitude periods of the signal
    :returns APQ
    """
    return PQ(PAS,11)

def PPQ(PPS):
    """
    Period perturbation quotient (PPQ)
    :params PPS: sequence of pitch periods of the signal
    :returns PPQ
    """
    return PQ(PPS,5)




import numpy as np
import scipy.stats as st
from sklearn.metrics import mean_squared_error


def F0feat(F0):

    F0nz=F0[F0!=0]
    tsf0=0.01
    if len(F0nz)==0:
        return np.zeros(30)

    F0mean=np.mean(F0nz)
    F0std=np.std(F0nz)
    F0max=np.max(F0nz)
    F0min=np.min(F0nz)
    F0skew=st.skew(F0nz)
    F0kurt=st.kurtosis(F0nz)
    tilt=[]
    mse=[]
    F0nzlist=np.split(F0, np.where(F0 == 0)[0]+1)
    F0nzlist=[F0nzlist[j] for j in range(len(F0nzlist) ) if len(F0nzlist[j])>1]
    F0nzlist=[F0nzlist[j][:-1] for j in  range(len(F0nzlist) )]
    for k in range(len(F0nzlist)):
        if len(F0nzlist[k])>1:
            t=np.arange(len(F0nzlist[k]))*tsf0
            pol=np.polyfit(t, F0nzlist[k],1)
            if not np.isnan(pol[0]):
                tilt.append(pol[0])
                f0rec=t*pol[0]+pol[1]
                msef0t=mean_squared_error(F0nzlist[k],f0rec)
                mse.append(msef0t)

    if len(tilt) == 0:
        tilt.append(0)            
        mse.append(0)
        
    tiltmean=np.mean(np.asarray(tilt))
    msemean=np.mean(np.asarray(mse))
    tiltstd=np.std(np.asarray(tilt))
    msestd=np.std(np.asarray(mse))
    tiltmax=np.max(np.asarray(tilt))
    msemax=np.max(np.asarray(mse))
    tiltmin=np.min(np.asarray(tilt))
    msemin=np.min(np.asarray(mse))
    tiltskw=st.skew(np.asarray(tilt))
    mseskw=st.skew(np.asarray(mse))
    tiltku=st.kurtosis(np.asarray(tilt))
    mseku=st.kurtosis(np.asarray(mse))

    f01mean=np.mean(F0nzlist[0])
    f01std=np.std(F0nzlist[0])
    f01max=np.max(F0nzlist[0])
    f01min=np.min(F0nzlist[0])
    f01skw=st.skew(F0nzlist[0])
    f01ku=st.kurtosis(F0nzlist[0])

    f0lmean=np.mean(F0nzlist[-1])
    f0lstd=np.std(F0nzlist[-1])
    f0lmax=np.max(F0nzlist[-1])
    f0lmin=np.min(F0nzlist[-1])
    f0lskw=st.skew(F0nzlist[-1])
    f0lku=st.kurtosis(F0nzlist[-1])
    featv=np.hstack([F0mean, F0std, F0max, F0min, F0skew, F0kurt, tiltmean, msemean, tiltstd, msestd, tiltmax, msemax, tiltmin, msemin,tiltskw, mseskw, tiltku, mseku, f01mean, f01std, f01max, f01min, f01skw, f01ku, f0lmean, f0lstd, f0lmax, f0lmin, f0lskw, f0lku])
    return featv



def energy_cont_segm(segments, size_frameS, size_stepS):

    E=[]
    for j in range(len(segments)):
        logE=[]
        segment_t=segments[j]
        overlap=size_stepS/size_frameS
        nF=int((len(segment_t)/size_frameS/overlap))-1
        if nF>1:
            for l in range(nF):
                data_frame=segment_t[int(l*size_stepS):int(l*size_stepS+size_frameS)]
                logE.append(logEnergy(data_frame))
            E.append(np.asarray(logE))
    return E



def polyf0(F0):
    tsf0=0.01

    F0rec=[]
    F0v=[]
    F0u=[]
    total=0
    for j in range(len(F0)-1):
        if F0[j]>0 and F0[j+1]==0:
            offset=True
            onset=False
            voiced=False
            Unvoiced=False
        elif F0[j]>0 and F0[j+1]>0:
            offset=False
            onset=False
            voiced=True
            Unvoiced=False
        elif F0[j]==0 and F0[j+1]==0:
            offset=False
            onset=False
            voiced=False
            Unvoiced=True
        elif F0[j]==0 and F0[j+1]>0:
            offset=False
            onset=True
            voiced=False
            Unvoiced=False

        if voiced:
            F0v.append(F0[j])
        elif Unvoiced:
            F0u.append(0)

        elif onset:
            F0v=[]
            F0u.append(0)
            total+=len(F0u)
            F0rec.append(F0u)
        elif offset:
            F0v.append(F0[j])
            F0u=[]
            t=np.arange(len(F0v))*tsf0
            if len(F0v)>1:
                pol=np.polyfit(t, F0v,1)
                f0rec=t*pol[0]+pol[1]
                F0rec.append(f0rec)
            else:
                F0rec.append(np.zeros(len(F0v)))
            total+=len(F0v)
        


    F0rec=np.hstack(F0rec)

    diffend=len(F0)-len(F0rec)

    f0fill=np.zeros(diffend)
    F0rec=np.hstack((F0rec, f0fill))

    return F0rec


def energy_feat(segments, fs, size_frameS, size_stepS):

    first_segment=0
    last_segment=len(segments)-1
    E_mean=[]
    E_std=[]
    E_skw=[]
    E_ku=[]
    tilt=[]
    mseE=[]
    if len(segments)<2:
        return np.zeros(24)

    for j in range(len(segments)):
        tsE=size_stepS/fs
        
        segment_t=segments[j]
        overlap=size_stepS/size_frameS
        log_energy_contour = get_energy_segment(size_frameS, size_stepS, segment_t, overlap)

        if j==first_segment and len(log_energy_contour)==0:
            first_segment+=1

        if j==last_segment and len(log_energy_contour)==0:
            last_segment-=1

        if len(log_energy_contour)<=1:
            continue
        E_mean.append(np.mean(log_energy_contour))
        E_std.append(np.std(log_energy_contour))
        E_skw.append(st.skew(log_energy_contour))
        E_ku.append(st.kurtosis(log_energy_contour))

        t=np.arange(len(log_energy_contour))*tsE
        pol=np.polyfit(t, log_energy_contour,1)
        if not np.isnan(pol[0]):
            tilt.append(pol[0])
            Erec=t*pol[0]+pol[1]
            mseE.append(mean_squared_error(log_energy_contour,Erec))
    
    E_mean=np.asarray(E_mean)
    E_std=np.asarray(E_std)
    E_skw=np.asarray(E_skw)
    E_ku=np.asarray(E_ku)
    tilt=np.asarray(tilt)
    mseE=np.asarray(mseE)
    if first_segment==len(segments):
        E0mean=0
        E0std=0
        E0max=0
        E0min=0
        E0skw=0
        E0ku=0
    else:
        segment_t=segments[first_segment]
        log_energy_contour = get_energy_segment(size_frameS, size_stepS, segment_t, overlap)
        E0mean=np.mean(log_energy_contour)
        E0std=np.std(log_energy_contour)
        E0max=np.max(log_energy_contour)
        E0min=np.min(log_energy_contour)
        E0skw=st.skew(log_energy_contour)
        E0ku=st.kurtosis(log_energy_contour)

    if last_segment==0:
        Elmean=0
        Elstd=0
        Elmax=0
        Elmin=0
        Elskw=0
        Elku=0
    else:
        segment_t=segments[last_segment]
        log_energy_contour = get_energy_segment(size_frameS, size_stepS, segment_t, overlap)

        if len(log_energy_contour)>0:
            Elmean=np.mean(log_energy_contour)
            Elstd=np.std(log_energy_contour)
            Elmax=np.max(log_energy_contour)
            Elmin=np.min(log_energy_contour)
            Elskw=st.skew(log_energy_contour)
            Elku=st.kurtosis(log_energy_contour)
        else:
            Elmean=0
            Elstd=0
            Elmax=0
            Elmin=0
            Elskw=0
            Elku=0

    feat_vec=np.hstack([np.mean(E_mean), np.mean(E_std), np.mean(E_skw), np.mean(E_ku), np.mean(tilt), np.std(tilt), st.skew(tilt), st.kurtosis(tilt), np.mean(mseE), np.std(mseE), st.skew(mseE), st.kurtosis(mseE), E0mean, E0std, E0max, E0min, E0skw, E0ku, Elmean, Elstd, Elmax, Elmin, Elskw, Elku])

    return feat_vec

def get_energy_segment(size_frameS, size_stepS, segment_t, overlap):
    nF=int((len(segment_t)/size_frameS/overlap))-1
    logE=[]
    for l in range(nF):
        data_frame=segment_t[int(l*size_stepS):int(l*size_stepS+size_frameS)]
        logE.append(logEnergy(data_frame))
    E=np.asarray(logE)
    return E



def dur_seg(segments, fs):
    dur=[]
    for j in range(len(segments)):
        dur.append(len(segments[j])/fs)
    dur=np.asarray(dur)
    if len(dur)>1:
        return np.asarray([np.mean(dur), np.std(dur), st.skew(dur), st.kurtosis(dur), np.max(dur), np.min(dur)])
    return np.zeros(6)




def duration_feat(segV, segU, segPause, signal, fs):
    Vrate=fs*float(len(segV))/len(signal)
    
    durV=dur_seg(segV, fs)
    durU=dur_seg(segU, fs)
    durP=dur_seg(segPause, fs)

    if durP[0]!=0 and durU[0]!=0 and durV[0]!=0: 
        PVU=durP[0]/(durV[0]+durU[0])
        PU=durP[0]/durU[0]
        UVU=durU[0]/(durV[0]+durU[0])
        VVU=durV[0]/(durV[0]+durU[0])
        VP=durV[0]/durP[0]
        UP=durU[0]/durP[0]
        dur_ratio=np.hstack([PVU, PU, UVU, VVU, VP, UP])
    else:
        dur_ratio=np.zeros(6)

    feat_vec=np.hstack((Vrate, durV, durU, durP, dur_ratio))

    return feat_vec





def V_UV(F0, data_audio, type_seg, size_stepS):
    if type_seg=="Voiced":
        pitch_seg = np.where(F0!=0)[0]
    elif type_seg=="Unvoiced":
        pitch_seg = np.where(F0==0)[0]
    dchange = np.diff(pitch_seg)
    change = np.where(dchange>1)[0]
    if len(pitch_seg)==0:
        return []
    init_seg= (pitch_seg[0]*size_stepS)+size_stepS
    segment=[]
    for indx in change:
        end_seg = (pitch_seg[indx]*size_stepS)+size_stepS#To compute energy
        seg = data_audio[int(init_seg):int(end_seg)]#To compute energy
        segment.append(seg)
        init_seg = (pitch_seg[indx+1]*size_stepS)+size_stepS#To compute energy

    return segment




def logEnergy(sig):
    if len(sig)==0:
        return -1e30
    sig2=np.power(sig,2)
    sumsig2=np.sum(np.abs(sig2))/len(sig2)

    logE=10*np.log10(sumsig2)
    if np.isnan(logE) or np.isinf(logE):
        logE=-1e30

    return logE







import numpy as np
import pysptk
import matplotlib.pyplot as plt
try:
    from .peakdetect import peakdetect
except:
    from peakdetect import peakdetect

from scipy.stats import pearsonr
from scipy.linalg import toeplitz
try:
    from scipy.signal.windows import medfilt, hann, filtfilt, blackman, hamming, buttord, butter, lfiltic, lfilter
except:
    from scipy.signal import medfilt, hann, filtfilt, blackman, hamming, buttord, butter, lfiltic, lfilter


def smooth(a,WSZ):
    # a: NumPy 1-D array containing the data to be smoothed
    # WSZ: smoothing window size needs, which must be odd number,
    # as in the original MATLAB implementation
    out0 = np.convolve(a,np.ones(WSZ,dtype=int),'valid')/WSZ
    r = np.arange(1,WSZ-1,2)
    start = np.cumsum(a[:WSZ-1])[::2]/r
    stop = (np.cumsum(a[:-WSZ:-1])[::2]/r)[::-1]
    return np.concatenate((  start , out0, stop  ))

def create_continuous_smooth_f0(F0,VUV,x):
    #Function to create and f0 contour which is interpolated over unvoiced
    # regions and is heavily smoothed with median and moving average filters.
    # Initial settings
    F0=np.asarray(F0)
    med_len=17
    sm_len=17
    F02=F0[:]
    f0nzp=np.where(F0>0)[0]
    f0nz=F0[f0nzp]
    if len(f0nz)<med_len:
        F0min=np.min(f0nz)
        F0max=np.max(f0nz)
    else:
        F0min=np.min(medfilt(f0nz,med_len))
        F0max=np.max(medfilt(f0nz,med_len))

    posmin=np.where(F0<F0min)[0]
    posmax=np.where(F0>F0max)[0]
    F02[posmin]=F0min
    F02[posmax]=F0max

    f0_inter=F02[:]

    N=len(F0)

    start=[]
    stop=[]
    initV=np.where(VUV==1)[0]

    for j in range(initV[0],N-1):
        if VUV[j]==0 and VUV[j+1]==1:
            stop.append(j)
        elif VUV[j]==1 and VUV[j+1]==0:
            start.append(j)
    n_seg=np.min([len(start),len(stop)])
    for j in range(n_seg):
        f0_int_cur=np.interp(np.arange(start[j], stop[j]), [start[j], stop[j]], [F0[start[j]],F0[stop[j]+1]])
        f0_inter[start[j]+1:stop[j]+1]=f0_int_cur
    f0_inter=smooth(medfilt(f0_inter,med_len),sm_len)
    f0_samp=np.interp(np.arange(len(x)),np.linspace(0,len(x),len(f0_inter)),f0_inter)

    return f0_inter, f0_samp


def GetLPCresidual(wave,L,shift,order=24,VUV=1):

    """
    Get the LPC residual signal
    Written originally by Thomas Drugman, TCTS Lab.

    Adapated to python by
    J. C. Vasquez-Correa
    Pattern recognition Lab, University of Erlangen-Nuremberg
    Faculty of Enginerring, University of Antiqouia,

    :param wave: array with the speech signal
    :param L: window length (samples) (typ.25ms)
    :param shift: window shift (samples) (typ.5ms)
    :param order: LPC order
    :param VUV: vector of voicing decisions (=0 if Unvoiced, =1 if Voiced)
    :returns res: LPC residual signal 
    """
    start=0
    stop=int(start+L)
    res=np.zeros(len(wave))
    n=0
    while stop<len(wave):

        if np.sum(VUV[start:stop])==len(VUV[start:stop]): # if it is avoiced segment
            segment=wave[start:stop]
            segment=segment*hann(len(segment))
            try:
                A=pysptk.sptk.lpc(segment, order)
                inv=filtfilt(A,1,segment)
                inv=inv*np.sqrt(np.sum(segment**2)/np.sum(inv**2))
                res[start:stop]=inv
            except:
                print("WARNING: LPCs cannot be extracted for the segment")
        start=int(start+shift)
        stop=int(stop+shift)
        n=n+1
    res=res/max(abs(res))
    return res



def RCVD_reson_GCI(res,fs,F0mean):

    # Function to use the resonator used in RCVD (creaky voice detection),
    # applied to the LP-residual signal and give output

    ## Configure resonator (using settings in RCVD)

    Phi=2*np.pi*1*F0mean/fs
    Rho=0.9 # Set to a narrow bandwidth
    rep=filtfilt([1., 0., 0.],[1., -2*Rho*np.cos(Phi), Rho**2],res, padtype = 'odd', padlen=9) # Filter forwards and backwards to have zero-phase
    y=rep/np.max(np.abs(rep))
    return y


def zeroPhaseHPFilt(x,fs,f_p,f_s):
    Rp=0.5
    Rs=40
    Wp=f_p/(fs/2.)
    Ws=f_s/(fs/2.)
    [n,Wn] = buttord(Wp,Ws,Rp,Rs)
    [b,a]=butter(n,Wn,'high')
    y = filtfilt(b,a,x, padtype = 'odd', padlen=3*(max(len(b),len(a))-1))
    return y


def get_MBS(x,fs,T0mean):
    # Obtain the mean-based signal
    MBS=np.zeros(len(x))
    halfL=int(1.6*T0mean[0]/2)

    StepExp=3
    Step=2**StepExp
    for m in range(halfL, len(x)-halfL, Step):
        if len(T0mean)==1:
            halfL=int(1.7*T0mean[0]/2)
        else:
            halfL=int(1.7*T0mean[m]/2)
        Blackwin=blackman(2*halfL)
        start=int(m-halfL)
        stop=int(m+halfL)
        if stop>len(x):
            break
        if start>0:
            vec=x[start:stop]*Blackwin
            MBS[m]=np.mean(vec)

    t=np.where(MBS!=0)[0]
    MBS=np.interp(np.arange(len(x)), t, MBS[t])
    MBS[np.isnan(MBS)]=0
    MBS=zeroPhaseHPFilt(MBS,fs, 70,10)
    MBS=MBS/max(abs(MBS))
    MBS=smooth(MBS,7)
    return MBS


def get_MBS_GCI_intervals(MBS,fs,T0mean,F0max=500):
    F0max=F0max*2
    T0max=int(fs/F0max)
    [max_peaks, min_peaks]=peakdetect(MBS,lookahead = T0max)

    idx=np.asarray([min_peaks[j][0] for j in range(len(min_peaks))])
    N=len(idx)
    search_rate=0.28
    search_left_rate=0.01
    interval=np.zeros((N,2))

    for n in range(N):
        if len(T0mean)>1:
            start=idx[n]-int(T0mean[idx[n]]*search_left_rate)
            stop=idx[n]+int(T0mean[idx[n]]*search_rate)
        else:
            start=idx[n]-int(T0mean*search_left_rate)
            stop=idx[n]+int(T0mean*search_rate)
        if start<1:
            start=1
        if stop>len(MBS) and start<len(MBS):
            stop=len(MBS)
        elif stop>len(MBS) and start>=len(MBS):
            break
        interval[n,0]=start
        interval[n,1]=stop
    return interval


def search_res_interval_peaks(res,interval,Ncand,VUV):
    #Function to search for Ncand peaks in the residual within each search
    #interval

    N=len(interval)
    GCI=np.zeros((N,Ncand))
    rel_amp=np.zeros((N,Ncand))
    GCI_cur=0
    for n in range(N):
        start=int(interval[n][0])
        stop=int(interval[n][1])
        if stop<=start or np.sum(VUV[start:stop])!=len(VUV[start:stop]):
            continue
        if stop-start<Ncand:
            GCI_cur=GCI_cur+start-1
            GCI[n,:]=GCI_cur
            rel_amp[n,:]=0
        else:
            amp=np.sort(res[start:stop])
            amp=amp[::-1]
            idx=np.argsort(res[start:stop])
            idx=idx[::-1]
            GCI_cur=idx[0:Ncand]+start-1
            GCI[n,:]=np.asarray(GCI_cur)

            if max(amp[0:Ncand])>0:
                rel_amp[n,:]=1-(amp[0:Ncand]/max(amp[0:Ncand]))
            else:
                rel_amp[n,:]=np.ones(Ncand)

    GCI=[GCI[j] for j in range(len(GCI)) if sum(GCI[j])>0]
    rel_amp=[rel_amp[j] for j in range(len(GCI)) if sum(GCI[j])>0]
    return GCI, rel_amp


def RESON_dyProg_mat(GCI_relAmp,GCI_N,F0mean,x,fs,trans_wgt,relAmp_wgt, plots=True):
    # Function to carry out dynamic programming method described in Ney (1989)
    # and used previously in the ESPS GCI detection algorithm. The method
    # considers target costs and transition costs which are accumulated in
    # order to select the `cheapest' path, considering previous context

    # USAGE: INPUT
    #        GCI_relAmp - target cost matrix with N rows (GCI candidates) by M
    #                     columns (mean based signal derived intervals).
    #        GCI_N      - matrix containing N by M candidate GCI locations (in
    #                     samples)
    #        F0_inter   - F0 values updated every sample point
    #        x          - speech signal
    #        fs         - sampling frequency
    #        trans_wgt  - transition cost weight
    #
    #        OUTPUT
    #        GCI        - estimated glottal closure instants (in samples)
    # =========================================================================
    # === FUNCTION CODED BY JOHN KANE AT THE PHONETICS LAB TRINITY COLLEGE ====
    # === DUBLIN. 25TH October 2011 ===========================================
    # =========================================================================

    # =========================================================================
    # === FUNCTION ADAPTED AND CODED IN PYTHON BY J. C. Vasquez-Correa
    #   AT THE PATTERN RECOGNITION LAB, UNIVERSITY OF ERLANGEN-NUREMBERG ====
    # === ERLANGEN, MAY, 2018 ===========================================
    # =========================================================================



    ## Initial settings

    GCI_relAmp=np.asarray(GCI_relAmp)
    relAmp_wgt=np.asarray(relAmp_wgt)
    cost = GCI_relAmp*relAmp_wgt
    #print(cost.shape)
    GCI_N=np.asarray(GCI_N)
    ncands=GCI_N.shape[1]
    nframe=GCI_N.shape[0]
    #print(ncands, nframe, cost.shape)
    prev=np.zeros((nframe,ncands))
    pulseLen = int(fs/F0mean)

    for n in range(1,nframe):
        costm = get_costm_matrix(GCI_N, x, trans_wgt, ncands, pulseLen, n)
        costm=costm+np.tile(cost[n-1,0:ncands],(ncands,1))
        costm=np.asarray(costm)
        costi=np.min(costm,0)
        previ=np.argmin(costm,0)
        cost[n,0:ncands]=cost[n,0:ncands]+costi
        prev[n,0:ncands]=previ

    best=np.zeros(n+1)
    best[n]=np.argmin(cost[n,0:ncands])



    for i in range(n-1,1,-1):

        best[i-1]=prev[i,int(best[i])]

    GCI_opt=np.asarray([GCI_N[n,int(best[n])] for n in range(nframe)])

    if plots:
        GCI_opt_norm=np.asarray([GCI_opt[n]-GCI_N[n,0] for n in range(nframe)])
        plt.subplot(211)
        plt.plot(x)
        plt.stem(GCI_N[:,0], -0.1*np.ones(len(GCI_N[:,0])), 'r')
        plt.stem(GCI_opt, -0.1*np.ones(len(GCI_opt)), 'k')
        plt.subplot(212)
        plt.plot(GCI_opt, GCI_opt_norm, 'bo')
        plt.show()
    return GCI_opt

def get_costm_matrix(GCI_N, x, trans_wgt, ncands, pulseLen, n):
    costm=np.zeros((ncands,ncands))

    for c in range(ncands):
            #Transitions TO states in current frame
        start=int(GCI_N[n,c]-pulseLen/2)
        stop=min([int(GCI_N[n,c]+pulseLen/2), len(x)])
        pulse_cur=x[start:stop]

        for p in range(ncands):
                #Transitions FROM states in previous frame
            start=max([1,int(GCI_N[n-1,p]-pulseLen/2)])
            stop=min([len(x),int(GCI_N[n-1,p]+pulseLen/2)])
            pulse_prev=x[start:stop]
            if len(pulse_cur)==0 or np.isnan(pulse_cur[0]) or np.isnan(pulse_prev[0]):
                continue
            if len(pulse_cur)!=len(pulse_prev):
                cor_cur=0
            else:
                cor_cur=pearsonr(pulse_cur,pulse_prev)[0]
            costm[p,c]=(1-np.abs(cor_cur))*trans_wgt
    return costm


def lpcauto(s,p):
    #LPCAUTO  performs autocorrelation LPC analysis [AR,E,K]=(S,P)
    #  Inputs:
    #     s(ns)   is the input signal
    #	   p       is the order (default: 12)
    #
    # Outputs:
    #          ar(nf,p+1) are the AR coefficients with ar(1) = 1
    #          e(nf)      is the energy in the residual.
    #                     sqrt(e) is often called the 'gain' of the filter.

    nf=1
    ar=np.zeros(p+1)
    ar[0]=1
    dd=s[:]
    nc=len(s)
    pp=min(p,nc)
    ww=np.hamming(nc)
    y=np.zeros(nc+p)
    wd=dd*ww
    y[0:nc]=wd
    z=np.zeros((nc,pp+1))
    for j in range(pp+1):
        z[:,j]=y[j:j+len(s)]
    rr=np.matmul(wd,z)
    rm=toeplitz(rr[0:pp])
    rk=np.linalg.matrix_rank(rm)
    if rk!=0:
        if rk<pp:
            rm=rm[0:rk,0:rk]

        ar[1:rk+1]=np.matmul(-rr[1:rk+1], np.linalg.inv(rm))
    e=np.dot(rr,ar)
    return ar, e


def calc_residual(x,x_lpc,ord_lpc,GCI):

    # Function to carry out LPC analysis and inverse filtering, used in IAIF.m
    # function.

    # USAGE:
    #       Input:
    #             x       : signal to be inverse filtered
    #             x_lpc   : signal to carry out LPC analysis on
    #             ord_lpc : LPC prediction order
    #             GCI     : Glottal closure instants (in samples)
    #
    #       Output:
    #             vector_res : residual after inverse filtering
    #
    #
    #########################################################################
    ## Function Coded by John Kane @ The Phonetics and Speech Lab ###########
    ## Trinity College Dublin, August 2012 ##################################
    #########################################################################

    #########################################################################
    ## Function transalated to Python by J. C. Vasquez-Correa @ University of Erlangen-Nuremberg ###########
    #########################################################################

    vector_res=np.zeros(len(x))
    ze_lpc=np.zeros(ord_lpc)
    ar_lpc=np.zeros((ord_lpc+1, len(GCI)))
    for n in range(len(GCI)-1):
        if n>1:
            T0_cur=GCI[n]-GCI[n-1]
        else:
            T0_cur=GCI[1]-GCI[0]

        if GCI[n]-T0_cur<=0 or GCI[n]+T0_cur>len(x) or T0_cur<=0:
            continue

        start=int(GCI[n]-T0_cur)
        stop=int(GCI[n]+T0_cur)
        frame_res=x[start:stop]

        frame_lpc=x_lpc[start:stop]
        if len(frame_lpc)>ord_lpc*1.5:
            frame_wind=frame_lpc*hamming(len(frame_lpc))
            ar,e=lpcauto(frame_wind, ord_lpc)
            ar_par=np.real(ar)
            ar_lpc[:,n]=ar_par

        # inverse filtering
        try:
            
            if n>1:
                last_input=frame_res[::-1]
                last_output=residual[::-1]
                ze_lpc=lfiltic(ar_par, np.sqrt(e), last_output, last_input)
            residual=lfilter(b=ar_par,a=np.sqrt(e), x=frame_res, zi=ze_lpc)
        
        except:
            residual=[frame_res]
        residual_win=residual[0]*hamming(len(residual[0]))
        try:
            vector_res[start:stop]=vector_res[start:stop]+residual_win
        except:
            vector_res[start:start+len(residual_win)]=vector_res[start:start+len(residual_win)]+residual_win

    return vector_res



import torch
from torch import nn
import torch.nn.functional as F



class RAEenc(nn.Module):
    def __init__(self, dim=32):
        super().__init__()
        self.lstm1=nn.LSTM(128,64, batch_first=True, bidirectional=True)
        self.linear = nn.Linear(256, dim)

    def forward(self, x):
        x=x[:,0,:,:]
        x=x.permute(0,2,1)
        x,(hn,cn)=self.lstm1(x)
        #print(hn.size())
        hn=hn.permute(1,0,2)
        #print(hn.size()) 
        x=x[:,-1,:]
        hn=hn.contiguous().view(hn.size(0),-1)
        x = x.view(x.size(0), -1)
        x2=torch.cat((x,hn),1)
        x = F.leaky_relu(self.linear(x2))
        
        
        return x

class RAEdec(nn.Module):
    def __init__(self, dim=32, seq_len=126):
        super().__init__()
        self.lstm = nn.LSTM(dim,128, batch_first=True, num_layers=2)
        self.seq_len=seq_len

    def forward(self, x):
        x = torch.cat([x] *  self.seq_len, 1).view(x.size(0), self.seq_len, x.size(1))
        x, (h,c)=self.lstm(x)
        x=x.permute(0,2,1)
        x=x.view(x.size(0), 1, x.size(1), x.size(2))
        return x


class RAEn(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.encoder = RAEenc(dim=dim)
        self.decoder = RAEdec(dim=dim)

    def forward(self, x):
        bottleneck = self.encoder(x)
        x = self.decoder(bottleneck)
        return x, bottleneck




import os
import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pysptk
import torch
from scipy.integrate import cumtrapz
from scipy.io.wavfile import read
from tqdm import tqdm

PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(PATH, '..'))
sys.path.append(PATH)
from script_mananger import script_manager
from utils import dynamic2static, get_dict, save_dict_kaldimat
plt.rcParams["font.family"] = "Times New Roman"
from GCI import iaif, se_vq_varf0, get_vq_params


class Glottal:
    """
    Compute features based on the glottal source reconstruction from sustained vowels and continuous speech.

    For continuous speech, the features are computed over voiced segments

    Nine descriptors are computed:

    1. Variability of time between consecutive glottal closure instants (GCI)
    2. Average opening quotient (OQ) for consecutive glottal cycles-> rate of opening phase duration / duration of glottal cycle
    3. Variability of opening quotient (OQ) for consecutive glottal cycles-> rate of opening phase duration /duration of glottal cycle
    4. Average normalized amplitude quotient (NAQ) for consecutive glottal cycles-> ratio of the amplitude quotient and the duration of the glottal cycle
    5. Variability of normalized amplitude quotient (NAQ) for consecutive glottal cycles-> ratio of the amplitude quotient and the duration of the glottal cycle
    6. Average H1H2: Difference between the first two harmonics of the glottal flow signal
    7. Variability H1H2: Difference between the first two harmonics of the glottal flow signal
    8. Average of Harmonic richness factor (HRF): ratio of the sum of the harmonics amplitude and the amplitude of the fundamental frequency
    9. Variability of HRF

    Static or dynamic matrices can be computed:

    Static matrix is formed with 36 features formed with (9 descriptors) x (4 functionals: mean, std, skewness, kurtosis)

    Dynamic matrix is formed with the 9 descriptors computed for frames of 200 ms length with a time-shift of 50 ms.

    Notes:

    1. The fundamental frequency is computed using the RAPT algorithm.

    >>> python glottal.py <file_or_folder_audio> <file_features> <dynamic_or_static> <plots (true,  false)> <format (csv, txt, npy, kaldi, torch)>

    Examples command line:

    >>> python glottal.py "../audios/001_a1_PCGITA.wav" "glottalfeaturesAst.txt" "static" "true" "txt"
    >>> python glottal.py "../audios/098_u1_PCGITA.wav" "glottalfeaturesUst.csv" "static" "true" "csv"
    >>> python glottal.py "../audios/098_u1_PCGITA.wav" "glottalfeaturesUst.ark" "dynamic" "true" "kaldi"
    >>> python glottal.py "../audios/098_u1_PCGITA.wav" "glottalfeaturesUst.pt" "dynamic" "true" "torch"


    Examples directly in Python

    >>> from disvoice.glottal import Glottal
    >>> glottal=Glottal()
    >>> file_audio="../audios/001_a1_PCGITA.wav"
    >>> features=glottal.extract_features_file(file_audio, static, plots=True, fmt="numpy")
    >>> features2=glottal.extract_features_file(file_audio, static, plots=True, fmt="dataframe")
    >>> features3=glottal.extract_features_file(file_audio, dynamic, plots=True, fmt="torch")

    >>> path_audios="../audios/"
    >>> features1=glottal.extract_features_path(path_audios, static, plots=False, fmt="numpy")
    >>> features2=glottal.extract_features_path(path_audios, static, plots=False, fmt="torch")
    >>> features3=glottal.extract_features_path(path_audios, static, plots=False, fmt="dataframe")
    """

    def __init__(self):
        self.size_frame = 0.2
        self.size_step = 0.05
        self.head_dyn = ["var GCI", "avg NAQ", "std NAQ", "avg QOQ",
                         "std QOQ", "avg H1H2", "std H1H2", "avg HRF", "std HRF"]
        self.head_st = []
        for k in ["global avg", "global std", "global skewness", "global kurtosis"]:
            for h in self.head_dyn:
                self.head_st.append(k+" "+h)

    def plot_glottal(self, data_audio, fs, GCI, glottal_flow, glottal_sig):
        """Plots of the glottal features

        :param data_audio: speech signal.
        :param fs: sampling frequency
        :param GCI: glottal closure instants
        :param glottal_flow: glottal flow
        :param glottal_sig: reconstructed glottal signal
        :returns: plots of the glottal features.
        """

        fig, ax = plt.subplots(3, sharex=True)
        t = np.arange(0, float(len(data_audio))/fs, 1.0/fs)
        if len(t) > len(data_audio):
            t = t[:len(data_audio)]
        elif len(t) < len(data_audio):
            data_audio = data_audio[:len(t)]
        ax[0].plot(t, data_audio, 'k')
        ax[0].set_ylabel('Amplitude', fontsize=12)
        ax[0].set_xlim([0, t[-1]])
        ax[0].grid(True)

        ax[1].plot(t, glottal_sig, color='k', linewidth=2.0,
                   label="Glottal flow signal")
        amGCI = [glottal_sig[int(k-2)] for k in GCI]

        GCI = GCI/fs
        ax[1].plot(GCI, amGCI, 'bo', alpha=0.5, markersize=8, label="GCI")
        GCId = np.diff(GCI)
        ax[1].set_ylabel("Glottal flow", fontsize=12)
        ax[1].text(t[2], -0.8, "Avg. time consecutive GCI:" +
                   str(np.round(np.mean(GCId)*1000, 2))+" ms")
        ax[1].text(t[2], -1.05, "Std. time consecutive GCI:" +
                   str(np.round(np.std(GCId)*1000, 2))+" ms")
        ax[1].set_xlabel('Time (s)', fontsize=12)
        ax[1].set_xlim([0, t[-1]])
        ax[1].set_ylim([-1.1, 1.1])
        ax[1].grid(True)

        ax[1].legend(ncol=2, loc=2)

        ax[2].plot(t, glottal_flow, color='k', linewidth=2.0)
        ax[2].set_ylabel("Glotal flow derivative", fontsize=12)
        ax[2].set_xlabel('Time (s)', fontsize=12)
        ax[2].set_xlim([0, t[-1]])
        ax[2].grid(True)

        plt.show()

    def extract_glottal_signal(self, x, fs):
        """Extract the glottal flow and the glottal flow derivative signals

        :param x: data from the speech signal.
        :param fs: sampling frequency
        :returns: glottal signal
        :returns: derivative  of the glottal signal
        :returns: glottal closure instants

        >>> from scipy.io.wavfile import read
        >>> glottal=Glottal()
        >>> file_audio="../audios/001_a1_PCGITA.wav"
        >>> fs, data_audio=read(audio)
        >>> glottal, g_iaif, GCIs=glottal.extract_glottal_signal(data_audio, fs)

        """
        winlen = int(0.025*fs)
        winshift = int(0.005*fs)
        x = x-np.mean(x)
        x = x/float(np.max(np.abs(x)))
        GCIs = se_vq_varf0(x, fs)
        g_iaif = np.zeros(len(x))
        glottal = np.zeros(len(x))

        if GCIs is None:
            sys.warn("not enought voiced segments were found to compute GCI")
            return glottal, g_iaif, GCIs

        start = 0
        stop = int(start+winlen)
        win = np.hanning(winlen)

        while stop <= len(x):

            x_frame = x[start:stop]
            pGCIt = np.where((GCIs > start) & (GCIs < stop))[0]
            GCIt = GCIs[pGCIt]-start

            g_iaif_f = iaif(x_frame, fs, GCIt)
            glottal_f = cumtrapz(g_iaif_f, dx=1/fs)
            glottal_f = np.hstack((glottal[start], glottal_f))
            g_iaif[start:stop] = g_iaif[start:stop]+g_iaif_f*win
            glottal[start:stop] = glottal[start:stop]+glottal_f*win
            start = start+winshift
            stop = start+winlen
        g_iaif = g_iaif-np.mean(g_iaif)
        g_iaif = g_iaif/max(abs(g_iaif))

        glottal = glottal-np.mean(glottal)
        glottal = glottal/max(abs(glottal))
        glottal = glottal-np.mean(glottal)
        glottal = glottal/max(abs(glottal))

        return glottal, g_iaif, GCIs

    def extract_features_file(self, audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the glottal features from an audio file

        :param audio: .wav audio file.
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldi features, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> glottal=Glottal()
        >>> file_audio="../audios/001_a1_PCGITA.wav"
        >>> features1=glottal.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
        >>> features2=glottal.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
        >>> features3=glottal.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
        >>> glottal.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")
        """

        if static and fmt=="kaldi":
            raise ValueError("Kaldi is only supported for dynamic features")

        if audio.find('.wav') == -1 and audio.find('.WAV') == -1:
            raise ValueError(audio+" is not a valid wav file")
        fs, data_audio = read(audio)

        if len(data_audio.shape)>1:
            data_audio = data_audio.mean(1)

        data_audio = data_audio-np.mean(data_audio)
        data_audio = data_audio/float(np.max(np.abs(data_audio)))
        size_frameS = self.size_frame*float(fs)
        size_stepS = self.size_step*float(fs)
        overlap = size_stepS/size_frameS
        nF = int((len(data_audio)/size_frameS/overlap))-1
        data_audiof = np.asarray(data_audio*(2**15), dtype=np.float32)
        f0 = pysptk.sptk.rapt(data_audiof, fs, int(
            0.01*fs), min=20, max=500, voice_bias=-0.2, otype='f0')
        sizef0 = int(self.size_frame/0.01)
        stepf0 = int(self.size_step/0.01)
        startf0 = 0
        stopf0 = sizef0

        glottal, g_iaif, GCI = self.extract_glottal_signal(data_audio, fs)

        if plots:
            self.plot_glottal(data_audio, fs, GCI, g_iaif, glottal)

        avgGCIt = np.zeros(nF)
        varGCIt = np.zeros(nF)
        avgNAQt = np.zeros(nF)
        varNAQt = np.zeros(nF)
        avgQOQt = np.zeros(nF)
        varQOQt = np.zeros(nF)
        avgH1H2t = np.zeros(nF)
        varH1H2t = np.zeros(nF)
        avgHRFt = np.zeros(nF)
        varHRFt = np.zeros(nF)
        rmwin = []
        for l in range(nF):
            init = int(l*size_stepS)
            endi = int(l*size_stepS+size_frameS)
            gframe = glottal[init:endi]
            dgframe = glottal[init:endi]
            pGCIt = np.where((GCI > init) & (GCI < endi))[0]
            gci_s = GCI[pGCIt]-init
            f0_frame = f0[startf0:stopf0]
            pf0framez = np.where(f0_frame != 0)[0]
            f0nzframe = f0_frame[pf0framez]
            if len(f0nzframe) < 5:
                startf0 = startf0+stepf0
                stopf0 = stopf0+stepf0
                rmwin.append(l)
                continue

            startf0 = startf0+stepf0
            stopf0 = stopf0+stepf0
            GCId = np.diff(gci_s)
            avgGCIt[l] = np.mean(GCId/fs)
            varGCIt[l] = np.std(GCId/fs)
            NAQ, QOQ, T1, T2, H1H2, HRF = get_vq_params(
                gframe, dgframe, fs, gci_s)
            avgNAQt[l] = np.mean(NAQ)
            varNAQt[l] = np.std(NAQ)
            avgQOQt[l] = np.mean(QOQ)
            varQOQt[l] = np.std(QOQ)
            avgH1H2t[l] = np.mean(H1H2)
            varH1H2t[l] = np.std(H1H2)
            avgHRFt[l] = np.mean(HRF)
            varHRFt[l] = np.std(HRF)

        if static and len(rmwin) > 0:
            varGCIt = np.delete(varGCIt, rmwin)
            avgNAQt = np.delete(avgNAQt, rmwin)
            varNAQt = np.delete(varNAQt, rmwin)
            avgQOQt = np.delete(avgQOQt, rmwin)
            varQOQt = np.delete(varQOQt, rmwin)
            avgH1H2t = np.delete(avgH1H2t, rmwin)
            varH1H2t = np.delete(varH1H2t, rmwin)
            avgHRFt = np.delete(avgHRFt, rmwin)
            varHRFt = np.delete(varHRFt, rmwin)

        feat = np.stack((varGCIt, avgNAQt, varNAQt, avgQOQt,
                        varQOQt, avgH1H2t, varH1H2t, avgHRFt, varHRFt), axis=1)
        if static:
            feat=dynamic2static(feat)
            feat=np.expand_dims(feat, axis=0)
            head = self.head_st
        else:
            head=self.head_dyn

        if fmt in ("npy", "txt"):
            return feat

        elif fmt in ("dataframe", "csv"):
            df = {}
            for e, k in enumerate(head):
                df[k] = feat[:, e]
            return pd.DataFrame(df)
        elif fmt == "torch":
            return torch.from_numpy(feat)
        elif fmt == "kaldi":
            name_all = audio.split('/')
            dictX = {name_all[-1]: feat}
            save_dict_kaldimat(dictX, kaldi_file)

    def extract_features_path(self, path_audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the glottal features for audios inside a path

        :param path_audio: directory with (.wav) audio files inside, sampled at 16 kHz
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldifeatures, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> glottal=Glottal()
        >>> path_audio="../audios/"
        >>> features1=glottal.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=glottal.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=glottal.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
        >>> glottal.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")
        """
        hf = os.listdir(path_audio)
        hf.sort()

        pbar = tqdm(range(len(hf)))
        ids = []

        Features = []
        for j in pbar:
            pbar.set_description("Processing %s" % hf[j])
            audio_file = path_audio+hf[j]
            feat = self.extract_features_file(
                audio_file, static=static, plots=plots, fmt="npy")
            Features.append(feat)
            if static:
                ids.append(hf[j])
            else:
                ids.append(np.repeat(hf[j], feat.shape[0]))

        Features = np.vstack(Features)
        ids = np.hstack(ids)

        return self.save_features(Features, ids, fmt, static, kaldi_file)

    def save_features(self, Features, ids, fmt, static, kaldi_file):
        if static:
            head = self.head_st
        else:
            head = self.head_dyn
        
        if fmt in ("npy", "txt"):
            return Features
        if fmt in ("dataframe", "csv"):
            df = {}
            for e, k in enumerate(head):
                df[k] = Features[:, e]
            df["id"] = ids
            return pd.DataFrame(df)
        if fmt == "torch":
            return torch.from_numpy(Features)
        if fmt == "kaldi":
            dictX = get_dict(Features, ids)
            save_dict_kaldimat(dictX, kaldi_file)


if __name__ == "__main__":

    if len(sys.argv) != 6:
        print("python glottal.py <file_or_folder_audio> <file_features> <static (true, false)> <plots (true,  false)> <format (csv, txt, npy, kaldi, torch)>")
        sys.exit()

    glottal_o = Glottal()
    script_manager(sys.argv, glottal_o)



from torch import nn
import torch
import torch.nn.functional as F
import torch


class CAEenc(nn.Module):
    def __init__(self, dim=256, nc=1):
        super().__init__()
        self.conv1=nn.Conv2d(nc, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.pool=nn.MaxPool2d((2, 2))
        self.conv2=nn.Conv2d(16, 32, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(32)
        self.conv3=nn.Conv2d(32, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(64)
        self.conv4=nn.Conv2d(64, 128, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn4 = nn.BatchNorm2d(128)

        self.linear = nn.Linear(128*8*7, dim)

    def forward(self, x):
        x =F.leaky_relu((self.bn1(self.pool(self.conv1(x)))))
        x =F.leaky_relu((self.bn2(self.pool(self.conv2(x)))))
        x =F.leaky_relu((self.bn3(self.pool(self.conv3(x)))))
        x =F.leaky_relu((self.bn4(self.pool(self.conv4(x)))))
        x = x.view(x.size(0), -1)
        x = self.linear(x)
        return x

class CAEdec(nn.Module):
    def __init__(self, dim=256, nc=1):
        super().__init__()
        self.conv1=nn.ConvTranspose2d(128, 64, kernel_size=3, stride=1, padding=(1,0), bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.conv2=nn.ConvTranspose2d(64, 32, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(32)
        self.conv3=nn.ConvTranspose2d(32, 16, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(16)
        self.conv4=nn.ConvTranspose2d(16, nc, kernel_size=3, stride=1, padding=1, bias=False)
        self.linear = nn.Linear(dim,128*8*7)

    def forward(self, x):
        x = self.linear(x)
        x = x.view(x.size(0), 128, 8, 7)
        x = F.interpolate(x, scale_factor=2)
        x =F.leaky_relu((self.bn1(self.conv1(x))))
        x = F.interpolate(x, scale_factor=2)
        x =F.leaky_relu((self.bn2(self.conv2(x))))
        x = F.interpolate(x, scale_factor=2)
        x =F.leaky_relu((self.bn3(self.conv3(x))))
        x = F.interpolate(x, scale_factor=2)
        x =torch.sigmoid((self.conv4(x)))
        return x[:,:,:,0:-2]


class CAEn(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.encoder = CAEenc(dim=dim)
        self.decoder = CAEdec(dim=dim)

    def forward(self, x):
        bottleneck = self.encoder(x)
        x = self.decoder(bottleneck)
        return x, bottleneck




# -*- coding: utf-8 -*-


from .utils import *
from .script_mananger import script_manager


from . import (glottal, phonation, phonological, prosody, replearning, articulation)
from .glottal.glottal import Glottal
from .phonation.phonation import Phonation
from .phonological.phonological import Phonological
from .prosody.prosody import Prosody
from .replearning.replearning import RepLearning
from .articulation.articulation import Articulation




__all__=['Glottal', 'Phonation', 'Articulation', 'Prosody', 'Phonological', 'RepLearning']



# -*- coding: utf-8 -*-
"""
Created on Jul 21 2017

@author: J. C. Vasquez-Correa
"""


import numpy as np
import scipy.stats as st
import kaldi_io



def dynamic2static(feat):

    me=np.mean(feat,0)

    std=np.std(feat,0)
    sk=st.skew(feat)
    ku=st.kurtosis(feat)

    return np.hstack((me,std,sk,ku))

def dynamic2statict(feat):

    me=[]
    std=[]
    sk=[]
    ku=[]
    for k in feat:
        me.append(np.mean(k,0))
        std.append(np.std(k,0))
        sk.append(st.skew(k))
        ku.append(st.kurtosis(k))
    return np.hstack((me,std,sk,ku))


def dynamic2statict_artic(feat):

    me=[]
    std=[]
    sk=[]
    ku=[]
    for k in feat:
        if k.shape[0]>1:
            me.append(np.mean(k,0))
            std.append(np.std(k,0))
            sk.append(st.skew(k))
            ku.append(st.kurtosis(k))
        elif k.shape[0]==1:
            me.append(k[0,:])
            std.append(np.zeros(k.shape[1]))
            sk.append(np.zeros(k.shape[1]))
            ku.append(np.zeros(k.shape[1]))
        else:
            me.append(np.zeros(k.shape[1]))
            std.append(np.zeros(k.shape[1]))
            sk.append(np.zeros(k.shape[1]))
            ku.append(np.zeros(k.shape[1]))

    return np.hstack((np.hstack(me),np.hstack(std),np.hstack(sk),np.hstack(ku)))




def get_dict(feat_mat, IDs):
    uniqueids=np.unique(IDs)
    df={}
    for k in uniqueids:
        p=np.where(IDs==k)[0]
        featid=feat_mat[p,:]
        df[str(k)]=featid
    return df

def save_dict_kaldimat(dict_feat, temp_file):
    ark_scp_output='ark:| copy-feats --compress=true ark:- ark,scp:'+temp_file+'.ark,'+temp_file+'.scp'
    with kaldi_io.open_or_fd(ark_scp_output,'wb') as f:
        for key,mat in dict_feat.items(): 
            kaldi_io.write_mat(f, mat, key=key)

def multi_find(s, r):
    s_len = len(s)
    r_len = len(r)
    _complete = []
    if s_len < r_len:
        return -1
    for i in range(s_len):
        # search for r in s until not enough characters are left
        if s[i:i + r_len] == r:
            _complete.append(i)
        else:
            i = i + 1
    return _complete


def fill_when_empty(array):
    if len(array) == 0:
        return np.zeros((0,1))
    return array



# -*- coding: utf-8 -*-
"""
Created on Jul 21 2017, Modified Apr 10 2018.

@author: J. C. Vasquez-Correa, T. Arias-Vergara, J. S. Guerrero
"""



import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
import torch
import pandas as pd
import pysptk
from matplotlib import cm
from scipy.io.wavfile import read
import os
import sys
PATH = os.path.dirname(os.path.realpath(__file__))

sys.path.append(PATH+'/../')
sys.path.append(PATH)

plt.rcParams["font.family"] = "Times New Roman"
from prosody_functions import V_UV, F0feat, energy_cont_segm, polyf0, energy_feat, dur_seg, duration_feat, get_energy_segment

from script_mananger import script_manager
from utils import save_dict_kaldimat, get_dict
import praat.praat_functions as praat_functions
class Prosody:
    """
    Compute prosody features from continuous speech based on duration, fundamental frequency and energy.
    Static or dynamic matrices can be computed:
    Static matrix is formed with 103 features and include

    1-6     F0-contour:                                                       Avg., Std., Max., Min., Skewness, Kurtosis

    7-12    Tilt of a linear estimation of F0 for each voiced segment:        Avg., Std., Max., Min., Skewness, Kurtosis

    13-18   MSE of a linear estimation of F0 for each voiced segment:         Avg., Std., Max., Min., Skewness, Kurtosis

    19-24   F0 on the first voiced segment:                                   Avg., Std., Max., Min., Skewness, Kurtosis

    25-30   F0 on the last voiced segment:                                    Avg., Std., Max., Min., Skewness, Kurtosis

    31-34   energy-contour for voiced segments:                               Avg., Std., Skewness, Kurtosis

    35-38   Tilt of a linear estimation of energy contour for V segments:     Avg., Std., Skewness, Kurtosis

    39-42   MSE of a linear estimation of energy contour for V segment:       Avg., Std., Skewness, Kurtosis

    43-48   energy on the first voiced segment:                               Avg., Std., Max., Min., Skewness, Kurtosis

    49-54   energy on the last voiced segment:                                Avg., Std., Max., Min., Skewness, Kurtosis

    55-58   energy-contour for unvoiced segments:                             Avg., Std., Skewness, Kurtosis

    59-62   Tilt of a linear estimation of energy contour for U segments:     Avg., Std., Skewness, Kurtosis

    63-66   MSE of a linear estimation of energy contour for U segments:      Avg., Std., Skewness, Kurtosis

    67-72   energy on the first unvoiced segment:                             Avg., Std., Max., Min., Skewness, Kurtosis

    73-78   energy on the last unvoiced segment:                              Avg., Std., Max., Min., Skewness, Kurtosis

    79      Voiced rate:                                                      Number of voiced segments per second

    80-85   Duration of Voiced:                                               Avg., Std., Max., Min., Skewness, Kurtosis

    86-91   Duration of Unvoiced:                                             Avg., Std., Max., Min., Skewness, Kurtosis

    92-97   Duration of Pauses:                                               Avg., Std., Max., Min., Skewness, Kurtosis

    98-103  Duration ratios:                                                 Pause/(Voiced+Unvoiced), Pause/Unvoiced, Unvoiced/(Voiced+Unvoiced),Voiced/(Voiced+Unvoiced), Voiced/Puase, Unvoiced/Pause

    Dynamic matrix is formed with 13 features computed for each voiced segment and contains


    1-6. Coefficients of 5-degree Lagrange polynomial to model F0 contour

    7-12. Coefficients of 5-degree Lagrange polynomial to model energy contour

    13. Duration of the voiced segment

    Dynamic prosody features are based on
    Najim Dehak, "Modeling Prosodic Features With Joint Factor Analysis for Speaker Verification", 2007

    Script is called as follows

    >>> python prosody.py <file_or_folder_audio> <file_features> <static (true or false)> <plots (true or false)> <format (csv, txt, npy, kaldi, torch)>

    Examples command line:

    >>> python Prosody.py "../audios/001_ddk1_PCGITA.wav" "prosodyfeaturesAst.txt" "true" "true" "txt"
    >>> python Prosody.py "../audios/001_ddk1_PCGITA.wav" "prosodyfeaturesUst.csv" "true" "true" "csv"
    >>> python prosody.py "../audios/001_ddk1_PCGITA.wav" "prosodyfeaturesUdyn.pt" "false" "true" "torch"

    >>> python Prosody.py "../audios/" "prosodyfeaturesst.txt" "true" "false" "txt"
    >>> python Prosody.py "../audios/" "prosodyfeaturesst.csv" "true" "false" "csv"
    >>> python Prosody.py "../audios/" "prosodyfeaturesdyn.pt" "false" "false" "torch"
    >>> python Prosody.py "../audios/" "prosodyfeaturesdyn.csv" "false" "false" "csv"

    Examples directly in Python

    >>> prosody=Prosody()
    >>> file_audio="../audios/001_ddk1_PCGITA.wav"
    >>> features1=prosody.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
    >>> features2=prosody.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
    >>> features3=prosody.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
    >>> prosody.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")

    >>> path_audio="../audios/"
    >>> features1=prosody.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
    >>> features2=prosody.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
    >>> features3=prosody.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
    >>> prosody.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")

    """

    def __init__(self):
        self.pitch_method = "rapt"
        self.size_frame = 0.02
        self.step = 0.01
        self.thr_len = 0.14
        self.minf0 = 60
        self.maxf0 = 350
        self.voice_bias = -0.2
        self.P = 5
        self.namefeatf0 = ["F0avg", "F0std", "F0max", "F0min",
                           "F0skew", "F0kurt", "F0tiltavg", "F0mseavg",
                           "F0tiltstd", "F0msestd", "F0tiltmax", "F0msemax",
                           "F0tiltmin", "F0msemin", "F0tiltskw", "F0mseskw",
                           "F0tiltku", "F0mseku", "1F0mean", "1F0std",
                           "1F0max", "1F0min", "1F0skw", "1F0ku", "lastF0avg",
                           "lastF0std", "lastF0max", "lastF0min", "lastF0skw", "lastF0ku"]
        self.namefeatEv = ["avgEvoiced", "stdEvoiced", "skwEvoiced", "kurtosisEvoiced",
                           "avgtiltEvoiced", "stdtiltEvoiced", "skwtiltEvoiced", "kurtosistiltEvoiced",
                           "avgmseEvoiced", "stdmseEvoiced", "skwmseEvoiced", "kurtosismseEvoiced",
                           "avg1Evoiced", "std1Evoiced", "max1Evoiced", "min1Evoiced", "skw1Evoiced",
                           "kurtosis1Evoiced", "avglastEvoiced", "stdlastEvoiced", "maxlastEvoiced",
                           "minlastEvoiced", "skwlastEvoiced",  "kurtosislastEvoiced"]
        self.namefeatEu = ["avgEunvoiced", "stdEunvoiced", "skwEunvoiced", "kurtosisEunvoiced",
                           "avgtiltEunvoiced", "stdtiltEunvoiced", "skwtiltEunvoiced", "kurtosistiltEunvoiced",
                           "avgmseEunvoiced", "stdmseEunvoiced", "skwmseEunvoiced", "kurtosismseEunvoiced",
                           "avg1Eunvoiced", "std1Eunvoiced", "max1Eunvoiced", "min1Eunvoiced", "skw1Eunvoiced",
                           "kurtosis1Eunvoiced", "avglastEunvoiced", "stdlastEunvoiced", "maxlastEunvoiced",
                           "minlastEunvoiced", "skwlastEunvoiced",  "kurtosislastEunvoiced"]

        self.namefeatdur = ["Vrate", "avgdurvoiced", "stddurvoiced", "skwdurvoiced", "kurtosisdurvoiced", "maxdurvoiced", "mindurvoiced",
                            "avgdurunvoiced", "stddurunvoiced", "skwdurunvoiced", "kurtosisdurunvoiced", "maxdurunvoiced", "mindurunvoiced",
                            "avgdurpause", "stddurpause", "skwdurpause", "kurtosisdurpause", "maxdurpause", "mindurpause",
                            "PVU", "PU", "UVU", "VVU", "VP", "UP"]
        self.head_st = self.namefeatf0+self.namefeatEv+self.namefeatEu+self.namefeatdur

        self.namef0d = ["f0coef"+str(i) for i in range(6)]
        self.nameEd = ["Ecoef"+str(i) for i in range(6)]
        self.head_dyn = self.namef0d+self.nameEd+["Voiced duration"]

    def plot_pros(self, data_audio, fs, F0, segmentsV, segmentsU, F0_features):
        """Plots of the prosody features

        :param data_audio: speech signal.
        :param fs: sampling frequency
        :param F0: contour of the fundamental frequency
        :param segmentsV: list with the voiced segments
        :param segmentsU: list with the unvoiced segments
        :param F0_features: vector with f0-based features
        :returns: plots of the prosody features.
        """
        plt.figure(figsize=(6, 6))
        plt.subplot(211)
        ax1 = plt.gca()
        t = np.arange(len(data_audio))/float(fs)
        colors = cm.get_cmap('Accent', 5)
        ax1.plot(t, data_audio, 'k', label="speech signal",
                 alpha=0.5, color=colors.colors[4])
        ax1.set_ylabel('Amplitude', fontsize=12)
        ax1.set_xlabel('Time (s)', fontsize=12)
        ax1.set_xlim([0, t[-1]])
        ax2 = ax1.twinx()
        fsp = len(F0)/t[-1]
        t2 = np.arange(len(F0))/fsp
        ax2.plot(
            t2, F0, color=colors.colors[0], linewidth=2, label=r"Real $F_0$", alpha=0.5)
        ax2.set_ylabel(r'$F_0$ (Hz)', color=colors.colors[0], fontsize=12)
        ax2.tick_params('y', colors=colors.colors[0])

        p0 = np.where(F0 != 0)[0]
        f0avg = np.nanmean(np.where(F0 != 0, F0, np.nan))
        f0std = np.std(F0[p0])

        ax2.plot([t2[0], t2[-1]], [f0avg, f0avg],
                 color=colors.colors[2], label=r"Avg. $F_0$")
        ax2.fill_between([t2[0], t2[-1]], y1=[f0avg+f0std, f0avg+f0std], y2=[f0avg-f0std,
                         f0avg-f0std], color=colors.colors[2], alpha=0.2, label=r"Avg. $F_0\pm$ SD.")
        F0rec = polyf0(F0)
        ax2.plot(t2, F0rec, label=r"estimated $F_0$",
                 c=colors.colors[1], linewidth=2.0)
        plt.text(t2[2], np.max(F0)-5, r"$F_0$ SD.=" +
                 str(np.round(f0std, 1))+" Hz")
        plt.text(t2[2], np.max(F0)-20, r"$F_0$ tilt.=" +
                 str(np.round(F0_features[6], 1))+" Hz")

        plt.legend(ncol=2, loc=8)

        plt.subplot(212)
        size_frameS = 0.02*float(fs)
        size_stepS = 0.01*float(fs)

        logE = energy_cont_segm([data_audio], size_frameS, size_stepS)
        Esp = len(logE[0])/t[-1]
        t2 = np.arange(len(logE[0]))/float(Esp)
        plt.plot(t2, logE[0], color='k', linewidth=2.0)
        plt.xlabel('Time (s)', fontsize=12)
        plt.ylabel('Energy (dB)', fontsize=12)
        plt.xlim([0, t[-1]])
        plt.grid(True)
        plt.tight_layout()
        plt.show()

        plt.figure(figsize=(6, 3))
        Ev = energy_cont_segm(segmentsV, size_frameS, size_stepS)
        Eu = energy_cont_segm(segmentsU, size_frameS, size_stepS)

        plt.plot([np.mean(Ev[j])
                 for j in range(len(Ev))], label="Voiced energy")
        plt.plot([np.mean(Eu[j])
                 for j in range(len(Eu))], label="Unvoiced energy")

        plt.xlabel("Number of segments")
        plt.ylabel("Energy (dB)")
        plt.legend()
        plt.grid()
        plt.tight_layout()
        plt.show()

    def extract_static_features(self, audio, plots, fmt):
        features = self.prosody_static(audio, plots)
        if fmt in ("npy", "txt"):
            return features
        elif fmt in ("dataframe", "csv"):
            df = {}
            for e, k in enumerate(self.head_st):
                df[k] = [features[e]]
            return pd.DataFrame(df)
        elif fmt == "torch":
            feat_t = torch.from_numpy(features)
            return feat_t
        elif fmt == "kaldi":
            raise ValueError("Kaldi is only supported for dynamic features")
        raise ValueError("format" + fmt+" is not supported")

    def extract_dynamic_features(self, audio, fmt, kaldi_file=""):
        features = self.prosody_dynamic(audio)
        if fmt in ("npy", "txt"):
            return features
        if fmt in ("dataframe", "csv"):
            df = {}
            for e, k in enumerate(self.head_dyn):
                df[k] = features[:, e]
            return pd.DataFrame(df)
        if fmt == "torch":
            feat_t = torch.from_numpy(features)
            return feat_t
        if fmt == "kaldi":
            name_all = audio.split('/')
            dictX = {name_all[-1]: features}
            save_dict_kaldimat(dictX, kaldi_file)
        else:
            raise ValueError("format" + fmt+" is not supported")

    def extract_features_file(self, audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the prosody features from an audio file

        :param audio: .wav audio file.
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldi features, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> prosody=Prosody()
        >>> file_audio="../audios/001_ddk1_PCGITA.wav"
        >>> features1=prosody.extract_features_file(file_audio, static=True, plots=True, fmt="npy")
        >>> features2=prosody.extract_features_file(file_audio, static=True, plots=True, fmt="dataframe")
        >>> features3=prosody.extract_features_file(file_audio, static=False, plots=True, fmt="torch")
        >>> prosody.extract_features_file(file_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test")
        """
        if static:
            return self.extract_static_features(audio, plots, fmt)

        else:
            return self.extract_dynamic_features(audio, fmt, kaldi_file)

    def prosody_static(self, audio, plots):
        """Extract the static prosody features from an audio file

        :param audio: .wav audio file.
        :param plots: timeshift to extract the features
        :returns: array with the 103 prosody features

        >>> prosody=Prosody()
        >>> file_audio="../audios/001_ddk1_PCGITA.wav"
        >>> features=prosody.prosody_static(file_audio, plots=True)

        """
        fs, data_audio = read(audio)

        if len(data_audio.shape)>1:
            data_audio = data_audio.mean(1)
        data_audio = data_audio-np.mean(data_audio)
        data_audio = data_audio/float(np.max(np.abs(data_audio)))
        size_frameS = self.size_frame*float(fs)
        size_stepS = self.step*float(fs)
        thr_len_pause = self.thr_len*float(fs)

        if self.pitch_method == 'praat':
            name_audio = audio.split('/')
            temp_uuid = 'prosody'+name_audio[-1][0:-4]
            if not os.path.exists(PATH+'/../tempfiles/'):
                os.makedirs(PATH+'/../tempfiles/')
            temp_filename_f0 = PATH+'/../tempfiles/tempF0'+temp_uuid+'.txt'
            temp_filename_vuv = PATH+'/../tempfiles/tempVUV'+temp_uuid+'.txt'
            praat_functions.praat_vuv(audio, temp_filename_f0, temp_filename_vuv,
                                      time_stepF0=self.step, minf0=self.minf0, maxf0=self.maxf0)

            F0, _ = praat_functions.decodeF0(
                temp_filename_f0, len(data_audio)/float(fs), self.step)
            os.remove(temp_filename_f0)
            os.remove(temp_filename_vuv)
        elif self.pitch_method == 'rapt':
            data_audiof = np.asarray(data_audio*(2**15), dtype=np.float32)
            F0 = pysptk.sptk.rapt(data_audiof, fs, int(
                size_stepS), min=self.minf0, max=self.maxf0, voice_bias=self.voice_bias, otype='f0')

        segmentsV = V_UV(F0, data_audio, type_seg="Voiced",
                         size_stepS=size_stepS)
        segmentsUP = V_UV(F0, data_audio, type_seg="Unvoiced",
                          size_stepS=size_stepS)

        segmentsP = []
        segmentsU = []
        for k in range(len(segmentsUP)):
            if (len(segmentsUP[k]) > thr_len_pause):
                segmentsP.append(segmentsUP[k])
            else:
                segmentsU.append(segmentsUP[k])

        F0_features = F0feat(F0)
        energy_featuresV = energy_feat(segmentsV, fs, size_frameS, size_stepS)
        energy_featuresU = energy_feat(segmentsU, fs, size_frameS, size_stepS)
        duration_features = duration_feat(
            segmentsV, segmentsU, segmentsP, data_audio, fs)

        if plots:
            self.plot_pros(data_audio, fs, F0, segmentsV,
                           segmentsU, F0_features)

        features = np.hstack(
            (F0_features, energy_featuresV, energy_featuresU, duration_features))

        return features

    def prosody_dynamic(self, audio):
        """Extract the dynamic prosody features from an audio file

        :param audio: .wav audio file.
        :returns: array (N,13) with the prosody features extracted from an audio file.  N= number of voiced segments

        >>> prosody=Prosody()
        >>> file_audio="../audios/001_ddk1_PCGITA.wav"
        >>> features=prosody.prosody_dynamic(file_audio)

        """
        fs, data_audio = read(audio)

        if len(data_audio.shape)>1:
            data_audio = data_audio.mean(1)
        data_audio = data_audio-np.mean(data_audio)
        data_audio = data_audio/float(np.max(np.abs(data_audio)))
        size_frameS = self.size_frame*float(fs)
        size_stepS = self.step*float(fs)
        overlap = size_stepS/size_frameS

        if self.pitch_method == 'praat':
            name_audio = audio.split('/')
            temp_uuid = 'prosody'+name_audio[-1][0:-4]
            if not os.path.exists(PATH+'/../tempfiles/'):
                os.makedirs(PATH+'/../tempfiles/')
            temp_filename_f0 = PATH+'/../tempfiles/tempF0'+temp_uuid+'.txt'
            temp_filename_vuv = PATH+'/../tempfiles/tempVUV'+temp_uuid+'.txt'
            praat_functions.praat_vuv(audio, temp_filename_f0, temp_filename_vuv,
                                      time_stepF0=self.step, minf0=self.minf0, maxf0=self.maxf0)

            F0, _ = praat_functions.decodeF0(
                temp_filename_f0, len(data_audio)/float(fs), self.step)
            os.remove(temp_filename_f0)
            os.remove(temp_filename_vuv)
        elif self.pitch_method == 'rapt':
            data_audiof = np.asarray(data_audio*(2**15), dtype=np.float32)
            F0 = pysptk.sptk.rapt(data_audiof, fs, int(
                size_stepS), min=self.minf0, max=self.maxf0, voice_bias=self.voice_bias, otype='f0')

        pitchON = np.where(F0 != 0)[0]
        dchange = np.diff(pitchON)
        change = np.where(dchange > 1)[0]
        iniV = pitchON[0]

        featvec = []
        iniVoiced = (pitchON[0]*size_stepS)+size_stepS
        seg_voiced = []
        f0v = []
        Ev = []
        for indx in change:
            finV = pitchON[indx]+1
            finVoiced = (pitchON[indx]*size_stepS)+size_stepS
            VoicedSeg = data_audio[int(iniVoiced):int(finVoiced)]
            temp = F0[iniV:finV]
            tempvec = []
            if len(VoicedSeg) > int(size_frameS):
                seg_voiced.append(VoicedSeg)
                dur = len(VoicedSeg)/float(fs)
                x = np.arange(0,len(temp))
                z = np.poly1d(np.polyfit(x,temp,self.P))
                f0v.append(temp)
                tempvec.extend(z.coeffs)
                temp=get_energy_segment(size_frameS, size_stepS, VoicedSeg, overlap)
                Ev.append(temp)
                x = np.arange(0, len(temp))
                z = np.poly1d(np.polyfit(x, temp, self.P))
                tempvec.extend(z.coeffs)
                tempvec.append(dur)
                featvec.append(tempvec)
            iniV = pitchON[indx+1]
            iniVoiced = (pitchON[indx+1]*size_stepS)+size_stepS

        # Add the last voiced segment
        finV = (pitchON[len(pitchON)-1])
        finVoiced = (pitchON[len(pitchON)-1]*size_stepS)+size_stepS
        VoicedSeg = data_audio[int(iniVoiced):int(finVoiced)]
        temp = F0[iniV:finV]
        tempvec = []

        if len(VoicedSeg) > int(size_frameS):
            # Compute duration
            dur = len(VoicedSeg)/float(fs)
            
            x = np.arange(0, len(temp))
            z = np.poly1d(np.polyfit(x, temp, self.P))
            tempvec.extend(z.coeffs)
            # Energy coefficients
            temp=get_energy_segment(size_frameS, size_stepS, VoicedSeg, overlap)
            x = np.arange(0, len(temp))
            z = np.poly1d(np.polyfit(x, temp, self.P))
            tempvec.extend(z.coeffs)
            tempvec.append(dur)
            # Compute duration
            featvec.append(tempvec)

        return np.asarray(featvec)

    def extract_features_path(self, path_audio, static=True, plots=False, fmt="npy", kaldi_file=""):
        """Extract the prosody features for audios inside a path

        :param path_audio: directory with (.wav) audio files inside, sampled at 16 kHz
        :param static: whether to compute and return statistic functionals over the feature matrix, or return the feature matrix computed over frames
        :param plots: timeshift to extract the features
        :param fmt: format to return the features (npy, dataframe, torch, kaldi)
        :param kaldi_file: file to store kaldifeatures, only valid when fmt=="kaldi"
        :returns: features computed from the audio file.

        >>> prosody=Prosody()
        >>> path_audio="../audios/"
        >>> features1=prosody.extract_features_path(path_audio, static=True, plots=False, fmt="npy")
        >>> features2=prosody.extract_features_path(path_audio, static=True, plots=False, fmt="csv")
        >>> features3=prosody.extract_features_path(path_audio, static=False, plots=True, fmt="torch")
        >>> prosody.extract_features_path(path_audio, static=False, plots=False, fmt="kaldi", kaldi_file="./test.ark")
        """

        hf = os.listdir(path_audio)
        hf.sort()

        pbar = tqdm(range(len(hf)))
        ids = []

        Features = []
        for j in pbar:
            pbar.set_description("Processing %s" % hf[j])
            audio_file = path_audio+hf[j]
            feat = self.extract_features_file(
                audio_file, static=static, plots=plots, fmt="npy")
            Features.append(feat)
            if static:
                ids.append(hf[j])
            else:
                ids.append(np.repeat(hf[j], feat.shape[0]))

        Features = np.vstack(Features)
        ids = np.hstack(ids)

        return self.save_features(Features, ids, fmt, static, kaldi_file)

    def save_features(self, Features, ids, fmt, static, kaldi_file):

        if fmt in ("npy", "txt"):
            return Features
        elif fmt in ("dataframe", "csv"):
            df = {}
            if static:
                head = self.head_st
            else:
                head = self.head_dyn
            for e, k in enumerate(head):
                df[k] = Features[:, e]

            df["id"] = ids
            return pd.DataFrame(df)
        elif fmt == "torch":
            return torch.from_numpy(Features)
        elif fmt == "kaldi":
            if static:
                raise ValueError(
                    "Kaldi is only supported for dynamic features")
            dictX = get_dict(Features, ids)
            save_dict_kaldimat(dictX, kaldi_file)
        else:
            raise ValueError(fmt+" is not supported")


if __name__ == "__main__":
    if len(sys.argv) != 6:
        print("python prosody.py <file_or_folder_audio> <file_features> <static (true, false)> <plots (true,  false)> <format (csv, txt, npy, kaldi, torch)>")
        sys.exit()

    prosody = Prosody()
    script_manager(sys.argv, prosody)




import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from sklearn import metrics
import itertools

def plot_confusion_matrix(y_true, y_pred, classes,
                          normalize=True,
                          cmap=plt.cm.Blues):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    plt.figure(figsize=(5,5))
    cm = metrics.confusion_matrix(y_true, y_pred)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(j, i, np.round(cm[i, j],2),
                 horizontalalignment="center",
                 color="white" if cm[i, j] > thresh else "black")
    plt.ylabel('Real class')
    plt.xlabel('Predicted class')
    plt.tight_layout()
    plt.show()

def plot_ROC(ytest, score_test):
    plt.figure(figsize=(5,5))
    
    plt.plot([0, 1], [0, 1], linestyle='--', lw=2, color='r',
         label='Chance', alpha=.8)
    fprs, tprs, thresholds = metrics.roc_curve(ytest, score_test)
    AUC=metrics.auc(fprs, tprs)
    plt.plot(fprs, tprs, color='k',
         label=r'Avg. ROC (AUC = %0.3f)' % (AUC),
         lw=2, alpha=.8)

    plt.xlim([-0.05, 1.05])
    plt.ylim([-0.05, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.legend(loc="lower right")
    plt.grid()
    plt.tight_layout()
    
    plt.show()

def plot_histogram(ytest, score_test, name_clases):
    plt.figure(figsize=(5,5))
    p1=np.where(np.hstack(ytest)==0)[0]
    p2=np.where(np.hstack(ytest)==1)[0]
    sns.distplot(np.hstack(score_test)[p1], label=name_clases[0])
    sns.distplot(np.hstack(score_test)[p2], label=name_clases[1], color="k")
    plt.grid()
    plt.legend()
    plt.xlabel("Classification threshold")
    plt.ylabel('Normalized count')
    plt.tight_layout()
    plt.show()


import sys
from scipy.io.wavfile import read
import matplotlib.pyplot as plt
import numpy as np

import pysptk
import os
from scipy.integrate import cumtrapz
PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.append(PATH)

from peakdetect import peakdetect
from utils_gci import create_continuous_smooth_f0, GetLPCresidual, get_MBS, get_MBS_GCI_intervals, search_res_interval_peaks, RESON_dyProg_mat, calc_residual


T0_num=3 #Number of local glottal pulses to be used for harmonic spectrum
min_harm_num=5
HRF_freq_max=5000 # Maximum frequency used for harmonic measurement
qoq_level=0.5 # threhold for QOQ estimation
F0min=20
F0max=500

def se_vq_varf0(x,fs, f0=None):
    """
    Function to extract GCIs using an adapted version of the SEDREAMS 
    algorithm which is optimised for non-modal voice qualities (SE-VQ). Ncand maximum
    peaks are selected from the LP-residual signal in the interval defined by
    the mean-based signal. 
    
    A dynamic programming algorithm is then used to select the optimal path of GCI locations. 
    Then a post-processing method, using the output of a resonator applied to the residual signal, is
    carried out to remove false positives occurring in creaky speech regions.
    
    Note that this method is slightly different from the standard SE-VQ
    algorithm as the mean based signal is calculated using a variable window
    length. 
    
    This is set using an f0 contour interpolated over unvoiced
    regions and heavily smoothed. This is particularly useful for speech
    involving large f0 excursions (i.e. very expressive speech).

    :param x:  speech signal (in samples)
    :param fs: sampling frequency (Hz)
    :param f0: f0 contour (optional), otherwise its computed  using the RAPT algorithm
    :returns: GCI Glottal closure instants (in samples)
    
    References:
          Kane, J., Gobl, C., (2013) `Evaluation of glottal closure instant 
          detection in a range of voice qualities', Speech Communication
          55(2), pp. 295-314.
    

    ORIGINAL FUNCTION WAS CODED BY JOHN KANE AT THE PHONETICS AND SPEECH LAB IN 
    TRINITY COLLEGE DUBLIN ON 2013.
    
    THE SEDREAMS FUNCTION WAS CODED BY THOMAS DRUGMAN OF THE UNIVERSITY OF MONS
   
    THE CODE WAS TRANSLATED TO PYTHON AND ADAPTED BY J. C. Vasquez-Correa
    AT PATTERN RECOGNITION LAB UNIVERSITY OF ERLANGEN NUREMBER- GERMANY
    AND UNIVERSTY OF ANTIOQUIA, COLOMBIA
    JCAMILO.VASQUEZ@UDEA.EDU.CO
    https//jcvasquezc.github.io
    """
    if f0 is None:
        f0 = []
    if len(f0)==0 or sum(f0)==0:
        size_stepS=0.01*fs
        voice_bias=-0.2
        x=x-np.mean(x)
        x=x/np.max(np.abs(x))
        data_audiof=np.asarray(x*(2**15), dtype=np.float32)
        f0=pysptk.sptk.rapt(data_audiof, fs, int(size_stepS), min=F0min, max=F0max, voice_bias=voice_bias, otype='f0')


    F0nz=np.where(f0>0)[0]
    F0mean=np.median(f0[F0nz])
    VUV=np.zeros(len(f0))
    VUV[F0nz]=1
    if F0mean<70:
        F0mean=80

    # Interpolate f0 over unvoiced regions and heavily smooth the contour
    ptos=np.linspace(0,len(x),len(VUV))
    VUV_inter=np.interp(np.arange(len(x)), ptos, VUV)
    VUV_inter[np.where(VUV_inter>0.5)[0]]=1
    VUV_inter[np.where(VUV_inter<=0.5)[0]]=0
    f0_int, f0_samp=create_continuous_smooth_f0(f0,VUV,x)
    T0mean = fs/f0_samp
    winLen = 25 # window length in ms
    winShift = 5 # window shift in ms
    LPC_ord = int((fs/1000)+2) # LPC order
    Ncand=5 # Number of candidate GCI residual peaks to be considered in the dynamic programming
    trans_wgt=1 # Transition cost weight
    relAmp_wgt=0.3 # Local cost weight

    #Calculate LP-residual and extract N maxima per mean-based signal determined intervals
    res = GetLPCresidual(x,winLen*fs/1000,winShift*fs/1000,LPC_ord, VUV_inter) # Get LP residual
    MBS = get_MBS(x,fs,T0mean) # Extract mean based signal
    interval = get_MBS_GCI_intervals(MBS,fs,T0mean,F0max) # Define search intervals
    [GCI_N,GCI_relAmp] = search_res_interval_peaks(res,interval,Ncand, VUV_inter) # Find residual peaks
    if len(np.asarray(GCI_N).shape) > 1:
        GCI = RESON_dyProg_mat(GCI_relAmp,GCI_N,F0mean,x,fs,trans_wgt,relAmp_wgt, plots=False) # Do dynamic programming
    else:
        GCI = None

    return GCI

def iaif(x,fs,GCI):
    """
    Function to carry out iterative and adaptive inverse filtering (Alku et al 1992).
    
    :param x: speech signal (in samples)
    :param fs: sampling frequency (in Hz)
    :param GCI: Glottal closure instants (in samples)
    :returns: glottal flow derivative estimate
    

    Function Coded by John Kane @ The Phonetics and Speech Lab
    Trinity College Dublin, August 2012


    THE CODE WAS TRANSLATED TO PYTHON AND ADAPTED BY J. C. Vasquez-Correa
    AT PATTERN RECOGNITION LAB UNIVERSITY OF ERLANGEN NUREMBER- GERMANY
    AND UNIVERSTY OF ANTIOQUIA, COLOMBIA
    JCAMILO.VASQUEZ@UDEA.EDU.CO
    https//jcvasquezc.github.io
    """

    p=int(fs/1000)+2 # LPC order

    x_filt=x-np.mean(x)
    x_filt=x_filt/max(abs(x_filt))

    # ------------------------------------------------
    # emphasise high-frequencies of speech signal (LPC order 1) - PART 2 & 3
    ord_lpc1=1
    x_emph=calc_residual(x_filt,x_filt,ord_lpc1,GCI)

    # ------------------------------------------------
    # first estimation of the glottal source derivative - PART 4 & 5
    ord_lpc2=p
    residual1=calc_residual(x_filt,x_emph,ord_lpc2,GCI)

    # integration of the glottal source derivative to calculate the glottal
    # source pulse - PART 6 (cancelling lip radiation)
    ug1=cumtrapz(residual1)
    # ------------------------------------------------
    # elimination of the source effect from the speech spectrum - PART 7 & 8

    ord_lpc3=4
    vt_signal=calc_residual(x_filt,ug1,ord_lpc3,GCI)

    # ------------------------------------------------
    # second estimation of the glottal source signal - PART 9 & 10

    ord_lpc4=p
    residual2=calc_residual(x_filt,vt_signal,ord_lpc4,GCI)
    return residual2



def get_vq_params(gf, gfd, fs, GCI):
    """
    Function to estimate the glottal parameters: NAQ, QOQ, H1-H2, and HRF

    This function can be used to estimate a range of conventional glottal
    source parameters often used in the literature. This includes: the
    normalized amplitude quotient (NAQ), the quasi-open quotient (QOQ), the
    difference in amplitude of the first two harmonics of the differentiated
    glottal source spectrum (H1-H2), and the harmonic richness factor (HRF)
    
    :param gf: [samples] [N] Glottal flow estimation
    :param gfd: [samples] [N] Glottal flow derivative estimation
    :param fs: [Hz] [1] sampling frequency
    :param GCI: [samples] [M] Glottal closure instants
    :returns: NAQ [s,samples] [Mx2] Normalised amplitude quotient
    :returns: QOQ[s,samples] [Mx2] Quasi-open quotient
    :returns: H1H2[s,dB] [Mx2] Difference in glottal harmonic amplitude
    :returns: HRF[s,samples] [Mx2] Harmonic richness factor
    
    References:
     [1] Alku, P., B ackstrom, T., and Vilkman, E. Normalized amplitude quotient for parameterization of the glottal flow. Journal of the Acoustical Society of America, 112(2):701-710, 2002.
     
     [2] Hacki, T. Klassifizierung von glottisdysfunktionen mit hilfe der elektroglottographie. Folia Phoniatrica, pages 43-48, 1989.
     
     [3] Alku, P., Strik, H., and Vilkman, E. Parabolic spectral parameter - A new method for quantification of the glottal flow. Speech Communication, 22(1):67-79, 1997.
     
     [4] Hanson, H. M. Glottal characteristics of female speakers: Acoustic correlates. Journal of the Acoustical Society of America, 10(1):466-481, 1997.
        
     [5] Childers, D. G. and Lee, C. K. Voice quality factors: Analysis, synthesis and perception. Journal of the Acoustical Society of  America, 90(5):2394-2410, 1991.
    
    Function Coded by John Kane @ The Phonetics and Speech Lab
    Trinity College Dublin, August 2012

    THE CODE WAS TRANSLATED TO PYTHON AND ADAPTED BY J. C. Vasquez-Correa
    AT PATTERN RECOGNITION LAB UNIVERSITY OF ERLANGEN NUREMBERGER- GERMANY
    AND UNIVERSTY OF ANTIOQUIA, COLOMBIA
    JCAMILO.VASQUEZ@UDEA.EDU.CO
    https//jcvasquezc.github.io
    """


    NAQ=np.zeros(len(GCI))
    QOQ=np.zeros(len(GCI))
    H1H2=np.zeros(len(GCI))
    HRF=np.zeros(len(GCI))
    T1=np.zeros(len(GCI))
    T2=np.zeros(len(GCI))
    glot_shift=np.round(0.5/1000*fs)
    
    if len(GCI) <= 1:
        sys.warn("not enough voiced segments were found to compute GCI")
        return NAQ, QOQ, T1, T2, H1H2, HRF
    start=0
    stop=int(GCI[0])
    T0=GCI[1]-GCI[0]

    for n in range(len(GCI)):
        # Get glottal pulse compensated for zero-line drift
        if n>0:
            start=int(GCI[n-1])
            stop=int(GCI[n])
            T0=GCI[n]-GCI[n-1]
            if T0==0 and n>=2:
                T0=GCI[n]-GCI[n-2]
                start=int(GCI[n-2])
        F0=fs/T0

        if T0<=0 or F0<=F0min or F0>=F0max:
            continue

        gf_comb=[gf[start], gf[stop]]
        line=0
        if start!=stop and len(gf_comb)>1:
            line=np.interp(np.arange(stop-start), np.linspace(0,stop-start,2), gf_comb)
        elif start!=stop and len(gf_comb)<=1:
            line=gf_comb
        gf_seg=gf[start:stop]
        gf_seg_comp=gf_seg-line
        f_ac=np.max(gf_seg_comp)
        Amid=f_ac*qoq_level
        max_idx=np.argmax(gf_seg_comp)
        T1[n],T2[n] = find_amid_t(gf_seg_comp,Amid,max_idx)

        if stop+glot_shift<=len(gfd):
            stop=int(stop+glot_shift)
        gfd_seg=gfd[start:stop]

        # get NAQ and QOQ
        d_peak=np.max(np.abs(gfd_seg))
        
        NAQ[n]=(f_ac/d_peak)/T0
        QOQ[n]=(T2[n]-T1[n])/T0
        #Get frame positions for H1-H2 parameter
        H1H2[n], HRF[n]=compute_h1h2_hrf_frame(GCI[n], T0, T0_num, gfd, F0, fs)


    return NAQ, QOQ, T1, T2, H1H2, HRF

def compute_h1h2_hrf_frame(GCIn, T0, T0_num, gfd, F0, fs):
    H1H2=0
    HRF=0

    if GCIn-int((T0*T0_num)/2)>0:
        f_start=int(GCIn-int((T0*T0_num)/2))
    else:
        f_start=0
    if GCIn+int((T0*T0_num)/2)<=len(gfd):
        f_stop=int(GCIn+int((T0*T0_num)/2))
    else:
        f_stop=len(gfd)
    f_frame=gfd[f_start:f_stop]
    f_win=f_frame*np.hamming(len(f_frame))
    f_spec=20*np.log10(np.abs(np.fft.fft(f_win, fs)))

    f_spec=f_spec[0:int(len(f_spec)/2)]
    # get H1-H2 and HRF
    [max_peaks, min_peaks]=peakdetect(f_spec,lookahead = int(T0))


    if len(max_peaks)==0:
        return 0, 0
    h_idx, h_amp=zip(*max_peaks)
    HRF_harm_num=np.fix(HRF_freq_max/F0)
    if len(h_idx)>=min_harm_num:
        temp1=np.arange(HRF_harm_num)*F0
        f0_idx=np.zeros(len(h_idx))
        for mp in range(len(h_idx)):

            temp2=h_idx[mp]-temp1
            temp2=np.abs(temp2)
            posmin=np.where(temp2==min(temp2))[0]
            if len(posmin)>1:
                posmin=posmin[0]

            if posmin<len(h_idx):
                f0_idx[mp]=posmin
            else:
                f0_idx[mp]=len(h_idx)-1

        f0_idx=[int(mm) for mm in f0_idx]

        H1H2=h_amp[f0_idx[0]]-h_amp[f0_idx[1]]
        harms=[h_amp[mm] for mm in f0_idx[1:]]
        HRF=sum(harms)/h_amp[f0_idx[0]]

    return H1H2, HRF



def find_amid_t(glot_adj, Amid, Tz):
    #Function to find the start and stop positions of the quasi-open phase.
    T1=0
    T2=0
    if Tz!=0:
        n=Tz
        while glot_adj[n]>Amid and n>2:
            n=n-1
        T1=n
        n=Tz
        while glot_adj[n] > Amid and n < len(glot_adj)-1:
            n=n+1
        T2=n
    return T1, T2

if __name__=="__main__":
    if len(sys.argv)!=2:
        print("python GCI.py <audio_file.wav>")
        sys.exit()
    else:
        audio=sys.argv[1]

    fsi, data_audio=read(audio)
    GCIi=se_vq_varf0(data_audio,fsi)
    print('Glottal inverse filtering using IAIF algorithm (Alku et al. 1992)')
    g_iaif=iaif(data_audio,fsi,GCIi)
    g_iaif=g_iaif-np.mean(g_iaif)
    g_iaif=g_iaif/max(abs(g_iaif))

    print(data_audio.dtype)
    data_audio=data_audio/max(abs(data_audio))
    plt.figure()
    plt.plot(data_audio)
    plt.plot(g_iaif)

    plt.show()

    glottal=cumtrapz(g_iaif)
    glottal=glottal-np.mean(glottal)
    glottal=glottal/max(abs(glottal))

    plt.figure()
    plt.plot(glottal)
    plt.show()
